from qlisp import (CR, CX, CZ, SWAP, A, B, BellPhiM, BellPhiP, BellPsiM,
                   BellPsiP, H, S, Sdag, SQiSWAP, T, Tdag, U, Unitary2Angles,
                   applySeq, draw, fSim, iSWAP, kak_decomposition, kak_vector,
                   make_immutable, measure, phiminus, phiplus, psiminus,
                   psiplus, regesterGateMatrix, rfUnitary, seq2mat, sigmaI,
                   sigmaM, sigmaP, sigmaX, sigmaY, sigmaZ,
                   synchronize_global_phase)
from qlispc import (COMMAND, FREE, NOTSET, PUSH, READ, SYNC, TRIG, WRITE,
                    compile, get_arch, libraries, mapping_qubits,
                    register_arch)
from qlispc.kernel_utils import qcompile
from wath import (Interval, Primes, Transmon, Z2probs, complex_amp_to_real,
                  effective_temperature, exception, find_axis_of_symmetry,
                  find_center_of_symmetry, find_cross_point, fit_circle,
                  fit_cosine, fit_k, fit_max, fit_peaks, fit_pole, getFTMatrix,
                  graph, inv_poly, lin_fit, point_in_ellipse, point_in_polygon,
                  point_on_ellipse, point_on_segment, poly_fit, probs2Z,
                  relative_delay_to_absolute, thermal_excitation, viterbi_hmm)
from waveforms import (D, chirp, const, cos, cosh, coshPulse, cosPulse, cut,
                       drag, drag_sin, drag_sinx, exp, function, gaussian,
                       general_cosine, hanning, interp, mixing, one, poly,
                       registerBaseFunc, registerDerivative, samplingPoints,
                       sign, sin, sinc, sinh, square, step, t, wave_eval, zero)

from .executor.analyze import manual_analysis
from .executor.registry import Registry
from .executor.storage import find_report
from .executor.storage import get_report_by_index as get_report
from .executor.template import VAR
from .executor.utils import debug_analyze
from .scan import Scan, get_record, load_record, lookup, lookup_list
from .version import __version__
from .visualization import autoplot



import inspect
from pathlib import Path

from ..cli.config import get_config_value
from .load import load_workflow


class Node:

    def __init__(self, name: str):
        self.name = name
        self.dependents = []


class Tree:

    def __init__(self):
        self.nodes = {}
        self.heads = []

    def add_node(self, node: str):
        self.nodes[node] = Node(node)


def dependent_tree(node: str, code_path: str | Path) -> dict[str, list[str]]:
    '''
    Returns a dict of nodes and their dependents.
    '''
    tree = {}
    for n in load_workflow(node, code_path).depends():
        tree[n] = dependent_tree(n, code_path)
    return tree


def workflow_template(workflow: str, deps: list[str]) -> str:
    return f"""
import numpy as np
from loguru import logger

from qulab import VAR
from qulab.typing import Report


# 多长时间应该检查一次校准实验，单位是秒。
__timeout__ = 7*24*3600

def depends():
    return {deps!r}


async def calibrate():
    logger.info(f"running {workflow} ...")

    # calibrate 是一个完整的校准实验，如power Rabi，Ramsey等。
    # 你需要足够的扫描点，以使得后续的 analyze 可以拟合出合适的参数。

    # 这里只是一个示例，实际上你需要在这里写上你的校准代码。
    x = np.linspace(0, 2*np.pi, 101)
    y = []
    for i in x:
        y.append(np.sin(i))

    logger.info(f"running {workflow} ... finished!")
    return x, y


async def analyze(report: Report, history: Report | None = None) -> Report:
    \"\"\"
    分析校准结果。

    report: Report
        本次校准实验的数据。
    history: Report | None
        上次校准实验数据和分析结果，如果有的话。
    \"\"\"
    import random

    # 这里添加你的分析过程，运行 calibrate 得到的数据，在 report.data 里
    # 你可以得到校准的结果，然后根据这个结果进行分析。
    x, y = report.data

    # 完整校准后的状态有两种：OK 和 Bad，分别对应校准成功和校准失败。
    # 校准失败是指出现坏数据，无法简单通过重新运行本次校准解决，需要
    # 检查前置步骤。
    report.state = random.choice(['OK', 'Bad'])

    # 参数是一个字典，包含了本次校准得到的参数，后续会更新到config表中。
    report.parameters = {{'gate.R.Q1.params.amp':1}}

    # 其他信息可以是任何可序列化的内容，你可以将你想要记录的信息放在这里。
    # 下次校准分析时，这些信息也会在 history 参数中一起传入，帮助你在下
    # 次分析时对比参考。
    report.other_infomation = {{}}

    return report


async def check():
    logger.info(f"checking {workflow} ...")

    # check 是一个快速检查实验，用于检查校准是否过时。
    # 你只需要少数扫描点，让后续的 check_analyze 知道参数是否漂移，数据
    # 坏没坏就够了，不要求拟合。

    # 这里只是一个示例，实际上你需要在这里写上你的检查代码。
    x = np.linspace(0, 2*np.pi, 5)
    y = []
    for i in x:
        y.append(np.sin(i))

    logger.info(f"checking {workflow} ... finished!")
    return x, y


async def check_analyze(report: Report, history: Report | None = None) -> Report:
    \"\"\"
    分析检查结果。

    report: Report
        本次检查实验的数据。
    history: Report | None
        上次检查实验数据和分析结果，如果有的话。
    \"\"\"
    import random

    # 这里添加你的分析过程，运行 check 得到的数据，在 report.data 里
    # 你可以得到校准的结果，然后根据这个结果进行分析。
    x, y = report.data

    # 状态有三种：Outdated, OK, Bad，分别对应过时、正常、坏数据。
    # Outdated 是指数据过时，即参数漂了，需要重新校准。
    # OK 是指数据正常，参数也没漂，不用重新校准。
    # Bad 是指数据坏了，无法校准，需要检查前置步骤。
    report.state = random.choice(['Outdated', 'OK', 'Bad'])

    return report


async def oracle(report: Report,
           history: Report | None = None,
           system_state: dict[str:str] | None = None) -> Report:
    \"\"\"
    谕示：指凭直觉或经验判断，改动某些配置，以期望下次校准成功。
    
    当校准失败时，根据 analyze 的结果，尝试改变某些配置再重新校准整个系统。
    比如通常我们在死活测不到 rabi 或能谱时，会换一个 idle bias 再试试。这
    里我们凭直觉设的那个 bias 值，就是一个谕示，可以通过 oracle 来设定。

    该函数代入的参数 report 是 analyze 函数的返回值。
    \"\"\"

    # report.oracle['Q0.bias'] = 0.1
    # report.oracle['Q1.bias'] = -0.03

    return report
"""


async def debug_analyze(
        report_index: int,
        code_path: str | Path = get_config_value('code', Path),
        data_path: str | Path = get_config_value('data', Path),
) -> None:
    from .storage import get_report_by_index

    report = get_report_by_index(report_index, data_path)
    if report is None:
        raise ValueError(f'Invalid report index: {report_index}')
    workflow = report.workflow
    wf = load_workflow(workflow, code_path)
    if wf is None:
        raise ValueError(f'Invalid workflow: {workflow}')
    if hasattr(wf, '__QULAB_TEMPLATE__'):
        template_mtime = (Path(code_path) /
                          wf.__QULAB_TEMPLATE__).stat().st_mtime
        if template_mtime > wf.__mtime__:
            for k in dir(wf):
                if k.startswith('__VAR_') and len(k) == len('__VAR_17fb4dde'):
                    var_dict = getattr(wf, k)
                    break
            else:
                var_dict = {}
            wf = load_workflow((wf.__QULAB_TEMPLATE__, workflow, var_dict),
                               code_path)

    report = wf.analyze(report, report.previous)
    if inspect.isawaitable(report):
        report = await report
    if hasattr(wf, 'plot'):
        if inspect.iscoroutinefunction(wf.plot):
            await wf.plot(report)
        else:
            wf.plot(report)
    return report



import inspect
from collections import deque


def _parse_frame(frame):
    ret = {}
    ret['source'] = inspect.getsource(frame)
    ret['name'] = frame.f_code.co_name
    ret['firstlineno'] = 1
    if ret['name'] != '<module>':
        argnames = frame.f_code.co_varnames[:frame.f_code.co_argcount +
                                            frame.f_code.co_kwonlyargcount]
        ret['name'] += '(' + ', '.join(argnames) + ')'
        ret['firstlineno'] = frame.f_code.co_firstlineno
    ret['filename'] = frame.f_code.co_filename
    return ret


def _parse_traceback(err):
    ret = []
    tb = err.__traceback__
    while tb is not None:
        frame = _parse_frame(tb.tb_frame)
        frame['lineno'] = tb.tb_lineno
        ret.append(frame)
        tb = tb.tb_next
    return ret


def _format_frame(frame):
    post_lines = -1
    lines = deque(maxlen=16)
    lines.append(f"{frame['filename']} in {frame['name']}")
    for n, line in enumerate(frame['source'].split('\n')):
        lno = n + frame['firstlineno']
        if lno == frame['lineno']:
            lines.append(f"->{lno:3d} {line}")
            post_lines = 0
        else:
            lines.append(f"  {lno:3d} {line}")
            if post_lines >= 0:
                post_lines += 1
        if post_lines >= 7:
            break
    return '\n'.join(lines)


def _format_traceback(err):
    frame_text = []
    for frame in _parse_traceback(err):
        frame_text.append(_format_frame(frame))
    traceback_text = '\n'.join(frame_text)
    args = list(err.args)
    args.append(traceback_text)
    err.args = tuple(args)
    return err


###############################################################
# RPC Exceptions
###############################################################


class RPCError(Exception):
    """
    RPC base exception.
    """


class RPCServerError(RPCError):
    """
    Server side error.
    """

    @classmethod
    def make(cls, exce):
        exce = _format_traceback(exce)
        args = [exce.__class__.__name__]
        args.extend(list(exce.args))
        return cls(*args)

    def _repr_markdown_(self):
        return '\n'.join([
            '```python',
            '---------------------------------------------------------------------------',
            f'RPCServerError({self.args[0]})               Server raise:{self.args[1]}',
            f'{self.args[2]}',
            '---------------------------------------------------------------------------',
            '```',
        ])


class RPCTimeout(RPCError):
    """
    Timeout.
    """



import pickle
import threading
import time

import pyperclip
import zmq

from .storage import Report

# 需要复制到 Notebook 的代码模板
clip_template = """
from qulab.executor.analyze import get_report as get_report_{server_port}

report, history = get_report_{server_port}("tcp://{server_address}:{server_port}")
# 在这里插入数据处理逻辑

{analysis_code}
"""

analysis_code = """
# report.state = 'OK'
# report.parameters = {}
# report.oracle = {}
# report.other_infomation = {}

report.parameters

"""


# ZeroMQ 服务线程，用于响应 Notebook 的请求
class ServerThread(threading.Thread):

    def __init__(self, data, timeout):
        super().__init__()
        self.data = data
        self.result = None
        self.port = 0
        self.timeout = timeout
        self.running = True
        self.finished = threading.Event()
        self.context = zmq.Context()

    def find_free_port(self):
        with zmq.Socket(self.context, zmq.REQ) as s:
            s.bind_to_random_port("tcp://*")
            self.port = int(
                s.getsockopt(zmq.LAST_ENDPOINT).decode().split(":")[-1])
            s.unbind(s.getsockopt(zmq.LAST_ENDPOINT))
        return self.port

    def run(self):
        self.port = self.find_free_port()
        socket = self.context.socket(zmq.REP)
        socket.bind(f"tcp://*:{self.port}")
        # 设置 recv 超时 1 秒
        socket.RCVTIMEO = 1000
        start_time = time.time()
        try:
            while self.running and (time.time() - start_time < self.timeout):
                try:
                    msg = socket.recv()
                except zmq.Again:
                    continue  # 超时后继续等待

                # Notebook 端请求数据
                if msg == b"GET":
                    socket.send(pickle.dumps(self.data))
                else:
                    # Notebook 端提交了处理结果
                    try:
                        self.result = pickle.loads(msg)
                    except Exception as e:
                        # 如果解析失败，也返回默认 ACK
                        self.result = None
                    socket.send(b"ACK")
                    self.running = False
                    break
        finally:
            socket.close()
            self.context.term()
            self.finished.set()

    def stop(self):
        self.running = False
        self.finished.set()

    def wait_for_result(self):
        self.finished.wait()
        return self.result


# 进入分析流程，启动服务并打印等待提示
def get_result_or_wait_until_timeout(report: Report, history: Report | None,
                                     timeout: float) -> Report:
    server = ServerThread((report, history), timeout)
    server.start()

    parameters = report.parameters
    oracle = report.oracle
    other_infomation = report.other_infomation
    state = report.state

    # 格式化代码模板
    code = clip_template.format(server_address="127.0.0.1",
                                server_port=server.port,
                                analysis_code=analysis_code)

    # 将代码复制到剪切板
    pyperclip.copy(code)

    # 同时打印到终端，防止误操作导致剪切板内容丢失
    print("请在 Jupyter Notebook 中运行下面这段代码：")
    print("-" * 60)
    print(code)
    print("-" * 60)
    print("等待 Notebook 提交处理结果，或等待超时（{} 秒）...".format(timeout))

    start_time = time.time()
    # 采用循环等待提交结果，间隔 0.5 秒检测一次
    while server.finished.wait(timeout=0.5) is False:
        elapsed = time.time() - start_time
        if elapsed >= timeout:
            # 超时后结束等待
            server.stop()
            break

    result = server.wait_for_result()
    if result is None:
        return (state, parameters, oracle, other_infomation, code)
    else:
        return result


def manual_analysis(report: Report, history=None, timeout=3600):
    try:
        (state, parameters, oracle, other_infomation,
         code) = get_result_or_wait_until_timeout(report, history, timeout)
        report.parameters = parameters
        report.oracle = oracle
        report.state = state
        report.other_infomation = other_infomation
    except Exception as e:
        pass
    return report


def get_report(address: str) -> Report:
    import IPython

    ipy = IPython.get_ipython()
    if ipy is None:
        raise RuntimeError("请在 Jupyter Notebook 中运行此函数。")
    ipy.set_next_input(("from qulab.executor.analyze import submit_report\n"
                        "# 处理完成后，提交结果\n"
                        f"# submit_report(report, \"{address}\")"),
                       replace=False)
    context = zmq.Context()
    sock = context.socket(zmq.REQ)
    sock.connect(address)
    # 请求数据
    sock.send(b"GET")
    report, history = pickle.loads(sock.recv())
    return report, history


def submit_report(report: Report, address: str):
    import IPython

    ipy = IPython.get_ipython()
    if ipy is None:
        raise RuntimeError("请在 Jupyter Notebook 中运行此函数。")

    code = ipy.user_ns['In'][-2]

    parameters = report.parameters
    oracle = report.oracle
    other_infomation = report.other_infomation
    state = report.state

    context = zmq.Context()
    sock = context.socket(zmq.REQ)
    sock.connect(address)
    # 提交处理后的结果
    sock.send(pickle.dumps(
        (state, parameters, oracle, other_infomation, code)))
    ack = sock.recv()
    print("结果已提交。")



import shlex
import subprocess
import sys
import time

import click


def run_detached(executable_path):
    """
    启动可执行文件并完全分离（优先用 tmux/screen），无需额外终端窗口
    支持 Windows、Linux 和 macOS
    """
    try:
        if sys.platform == 'win32' or not _unix_detach_with_tmux_or_screen(
                executable_path):
            # 回退到带终端窗口的方案
            run_detached_with_terminal(executable_path)

    except Exception as e:
        click.echo(f"启动失败: {e}")
        sys.exit(1)


def _windows_start(executable_path):
    """Windows 弹窗启动方案"""
    subprocess.Popen(f'start cmd /k "{executable_path}"',
                     shell=True,
                     creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)


def _unix_detach_with_tmux_or_screen(executable_path):
    """Unix 后台分离方案（无窗口）"""
    safe_path = shlex.quote(executable_path)
    session_name = f"qulab_{int(time.time())}"

    # 尝试 tmux
    if _check_command_exists("tmux"):
        command = [
            "tmux",
            "new-session",
            "-d",
            "-s",
            session_name,
            executable_path + " ; tmux wait-for -S finished",  # 等待命令结束
            ";",
            "tmux",
            "wait-for",
            "finished"  # 防止进程立即退出
        ]
        subprocess.Popen(" ".join(command), shell=True, start_new_session=True)
        click.echo(f"已启动 tmux 会话: {session_name}")
        click.echo(f"你可以使用 `tmux attach -t {session_name}` 来查看输出")
        return True

    # 尝试 screen
    elif _check_command_exists("screen"):
        command = ["screen", "-dmS", session_name, executable_path]
        subprocess.Popen(command, start_new_session=True)
        click.echo(f"已启动 screen 会话: {session_name}")
        click.echo(f"你可以使用 `screen -r {session_name}` 来查看输出")
        return True

    return False


def run_detached_with_terminal(executable_path):
    """回退到带终端窗口的方案"""
    if sys.platform == 'win32':
        _windows_start(executable_path)
    elif sys.platform == 'darwin':
        script = f'tell app "Terminal" to do script "{executable_path}"'
        subprocess.Popen(["osascript", "-e", script], start_new_session=True)
    else:
        try:
            subprocess.Popen(
                ["gnome-terminal", "--", "sh", "-c", executable_path],
                start_new_session=True)
        except FileNotFoundError:
            subprocess.Popen(["xterm", "-e", executable_path],
                             start_new_session=True)


def _check_command_exists(cmd):
    """检查命令行工具是否存在"""
    try:
        subprocess.check_output(["which", cmd], stderr=subprocess.DEVNULL)
        return True
    except:
        return False


# 示例用法
if __name__ == '__main__':
    run_detached("/path/to/your/program")



import importlib
import os
import sys
from typing import Any

from ..cli.config import get_config_value
from .storage import Report, save_item

__current_config_id = None


def _query_config(name: str, default=None):
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    return parameters.get(name, default)


def _update_config(updates):
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    for k, v in updates.items():
        parameters[k] = v

    with open('parameters.pkl', 'wb') as f:
        pickle.dump(parameters, f)


def _delete_config(name: str):
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    if name in parameters:
        del parameters[name]

    with open('parameters.pkl', 'wb') as f:
        pickle.dump(parameters, f)


def _export_config() -> dict:
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    return parameters


def _clear_config():
    import pickle

    try:
        with open('parameters.pkl', 'rb') as f:
            parameters = pickle.load(f)
    except:
        parameters = {}

    parameters.clear()

    with open('parameters.pkl', 'wb') as f:
        pickle.dump(parameters, f)


def obey_the_oracle(report: Report, data_path):
    global __current_config_id
    update_config(report.oracle)
    cfg = export_config()
    __current_config_id = save_item(cfg, 'items', data_path)


def update_parameters(report: Report, data_path):
    global __current_config_id
    update_config(report.parameters)
    cfg = export_config()
    __current_config_id = save_item(cfg, 'items', data_path)


def current_config(data_path):
    global __current_config_id
    if __current_config_id is None:
        cfg = export_config()
        __current_config_id = save_item(cfg, 'items', data_path)
    return __current_config_id


query_config = _query_config
update_config = _update_config
delete_config = _delete_config
export_config = _export_config
clear_config = _clear_config


def set_config_api(query_method,
                   update_method,
                   delete_method,
                   export_method,
                   clear_method=None):
    """
    Set the query and update methods for the config.

    Args:
        query_method: The query method.
            the method should take a key and return the value.
        update_method: The update method.
            the method should take a dict of updates.
        delete_method: The delete method.
            the method should take a key and delete it.
        export_method: The export method.
            the method should return a dict of the config.
        clear_method: The clear method.
            the method should clear the config.
    """
    global query_config, update_config, delete_config, export_config, clear_config, _api

    query_config = query_method
    update_config = update_method
    delete_config = delete_method
    export_config = export_method
    clear_config = clear_method

    return query_config, update_config, delete_config, export_config, clear_config


def _init():
    code = get_config_value("code", str, default=None)
    if code is not None:
        code = os.path.expanduser(code)
        if code not in sys.path:
            sys.path.insert(0, code)

    api = get_config_value('api', str, None)
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)


_init()


class Registry():

    def __init__(self):
        self.api = (query_config, update_config, delete_config, export_config,
                    clear_config)

    def query(self, key: str) -> Any:
        return self.api[0](key)

    def get(self, key: str) -> Any:
        return self.query(key)

    def export(self) -> dict[str, dict[str, Any]]:
        return self.api[3]()

    def set(self, key: str, value: Any):
        return self.api[1]({key: value})

    def delete(self, key: str):
        return self.api[2](key)

    def clear(self):
        return self.api[4]()

    def update(self, parameters: dict[str, Any]):
        return self.api[1](parameters)



import asyncio
import contextlib
import copy
import inspect
import itertools
import lzma
import os
import pickle
import re
import sys
import uuid
from concurrent.futures import ProcessPoolExecutor
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Any, Awaitable, Callable, Iterable

import dill
import numpy as np
import zmq
from qlispreg.expression import Env, Expression, Symbol

from ..sys.rpc.zmq_socket import ZMQContextManager
from .optimize import NgOptimizer
from .record import Record
from .server import default_record_port
from .space import Optimizer, OptimizeSpace, Space
from .utils import (async_zip, call_function, dump_dict, dump_globals,
                    get_installed_packages, get_system_info, yapf_reformat)

try:
    from tqdm.notebook import tqdm
except:

    class tqdm():

        def update(self, n):
            pass

        def close(self):
            pass

        def reset(self):
            pass


__process_uuid = uuid.uuid1()
__task_counter = itertools.count()
__notebook_id = None

if os.getenv('QULAB_SERVER'):
    default_server = os.getenv('QULAB_SERVER')
else:
    default_server = f'tcp://127.0.0.1:{default_record_port}'
if os.getenv('QULAB_EXECUTOR'):
    default_executor = os.getenv('QULAB_EXECUTOR')
else:
    default_executor = default_server


class Promise():
    __slots__ = ['task', 'key', 'attr']

    def __init__(self, task, key=None, attr=None):
        self.task = task
        self.key = key
        self.attr = attr

    def __await__(self):

        async def _getitem(task, key):
            return (await task)[key]

        async def _getattr(task, attr):
            return getattr(await task, attr)

        if self.key is not None:
            return _getitem(self.task, self.key).__await__()
        elif self.attr is not None:
            return _getattr(self.task, self.attr).__await__()
        else:
            return self.task.__await__()

    def __getitem__(self, key):
        return Promise(self.task, key, None)

    def __getattr__(self, attr):
        return Promise(self.task, None, attr)


def current_notebook():
    return __notebook_id


async def create_notebook(name: str, database=default_server, socket=None):
    global __notebook_id

    async with ZMQContextManager(zmq.DEALER, connect=database,
                                 socket=socket) as socket:
        await socket.send_pyobj({'method': 'notebook_create', 'name': name})
        __notebook_id = await socket.recv_pyobj()


async def save_input_cells(notebook_id,
                           input_cells,
                           database=default_server,
                           socket=None):
    async with ZMQContextManager(zmq.DEALER, connect=database,
                                 socket=socket) as socket:
        await socket.send_pyobj({
            'method': 'notebook_extend',
            'notebook_id': notebook_id,
            'input_cells': input_cells
        })
        return await socket.recv_pyobj()


async def create_config(config: dict, database=default_server, socket=None):
    async with ZMQContextManager(zmq.DEALER, connect=database,
                                 socket=socket) as socket:
        buf = lzma.compress(pickle.dumps(config))
        await socket.send_pyobj({'method': 'config_update', 'update': buf})
        return await socket.recv_pyobj()


def task_uuid():
    return uuid.uuid3(__process_uuid, str(next(__task_counter)))


def _get_depends(func: Callable):
    try:
        sig = inspect.signature(func)
    except:
        return []

    args = []
    for name, param in sig.parameters.items():
        if param.kind in [
                param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD,
                param.KEYWORD_ONLY
        ]:
            args.append(name)
        elif param.kind == param.VAR_KEYWORD:
            pass
        elif param.kind == param.VAR_POSITIONAL:
            raise ValueError('not support VAR_POSITIONAL')
    return args


def _run_function_in_process(buf):
    func, args, kwds = dill.loads(buf)
    return func(*args, **kwds)


async def update_variables(variables: dict[str, Any], updates: dict[str, Any],
                           setters: dict[str, Callable]):
    coros = []
    for name, value in updates.items():
        if name in setters:
            coro = setters[name](value)
            if inspect.isawaitable(coro):
                coros.append(coro)
        variables[name] = value
    if coros:
        await asyncio.gather(*coros)


async def _iter_level(variables,
                      iters: list[tuple[str, Iterable | Expression | Callable
                                        | OptimizeSpace]],
                      order: list[list[str]],
                      functions: dict[str, Callable | Expression],
                      optimizers: dict[str, Optimizer],
                      setters: dict[str, Callable] = {},
                      getters: dict[str, Callable] = {}):
    iters_d = {}
    env = Env()
    env.variables = variables
    opts = {}

    for name, iter in iters:
        if isinstance(iter, OptimizeSpace):
            if iter.optimizer.name not in opts:
                opts[iter.optimizer.name] = iter.optimizer.create()
        elif isinstance(iter, Expression):
            iters_d[name] = iter.eval(env)
        elif isinstance(iter, Space):
            iters_d[name] = iter.toarray()
        elif callable(iter):
            iters_d[name] = await call_function(iter, variables)
        else:
            iters_d[name] = iter

    maxiter = 0xffffffff
    for name, opt in opts.items():
        opt_cfg = optimizers[name]
        maxiter = min(maxiter, opt_cfg.maxiter)

    async for args in async_zip(*iters_d.values(), range(maxiter)):
        await update_variables(variables, dict(zip(iters_d.keys(), args[:-1])),
                               setters)
        for name, opt in opts.items():
            args = opt.ask()
            opt_cfg = optimizers[name]
            await update_variables(variables, {
                n: v
                for n, v in zip(opt_cfg.dimensions.keys(), args)
            }, setters)

        await update_variables(
            variables, await call_many_functions(order, functions, variables),
            setters)

        yield variables

        variables.update(await call_many_functions(order, getters, variables))

        if opts:
            for key in list(variables.keys()):
                if key.startswith('*') or ',' in key:
                    await _unpack(key, variables)

        for name, opt in opts.items():
            opt_cfg = optimizers[name]
            args = [variables[n] for n in opt_cfg.dimensions.keys()]

            if name not in variables:
                raise ValueError(f'{name} not in variables.')
            fun = variables[name]
            if inspect.isawaitable(fun):
                fun = await fun
            if opt_cfg.minimize:
                opt.tell(args, fun)
            else:
                opt.tell(args, -fun)

    if opts:
        for name, opt in opts.items():
            opt_cfg = optimizers[name]
            result = opt.get_result()
            await update_variables(
                variables, {
                    name: value
                    for name, value in zip(opt_cfg.dimensions.keys(), result.x)
                }, setters)

        yield variables

        variables.update(await call_many_functions(order, getters, variables))

        for key in list(variables.keys()):
            if key.startswith('*') or ',' in key:
                await _unpack(key, variables)


async def call_many_functions(order: list[list[str]],
                              functions: dict[str, Callable],
                              variables: dict[str, Any]) -> dict[str, Any]:
    ret = {}
    for group in order:
        waited = []
        coros = []
        for name in group:
            if name in functions:
                waited.append(name)
                coros.append(call_function(functions[name], variables | ret))
        if coros:
            results = await asyncio.gather(*coros)
            ret.update(dict(zip(waited, results)))
    return ret


async def _unpack(key, variables):
    x = variables[key]
    if inspect.isawaitable(x):
        x = await x
    if key.startswith('**'):
        assert isinstance(
            x, dict), f"Should promise a dict for `**` symbol. {key}"
        if "{key}" in key:
            for k, v in x.items():
                variables[key[2:].format(key=k)] = v
        else:
            variables.update(x)
    elif key.startswith('*'):
        assert isinstance(
            x, (list, tuple,
                np.ndarray)), f"Should promise a list for `*` symbol. {key}"
        for i, v in enumerate(x):
            k = key[1:].format(i=i)
            variables[k] = v
    elif ',' in key:
        keys1, keys2 = [], []
        args = None
        for k in key.split(','):
            if k.startswith('*'):
                if args is None:
                    args = k
                else:
                    raise ValueError(f'Only one `*` symbol is allowed. {key}')
            elif args is None:
                keys1.append(k)
            else:
                keys2.append(k)
        assert isinstance(
            x,
            (list, tuple,
             np.ndarray)), f"Should promise a list for multiple symbols. {key}"
        if args is None:
            assert len(keys1) == len(
                x), f"Length of keys and values should be equal. {key}"
            for k, v in zip(keys1, x):
                variables[k] = v
        else:
            assert len(keys1) + len(keys2) <= len(
                x), f"Too many values for unpacking. {key}"
            for k, v in zip(keys1, x[:len(keys1)]):
                variables[k] = v
            end = -len(keys2) if keys2 else None
            for i, v in enumerate(x[len(keys1):end]):
                k = args[1:].format(i=i)
                variables[k] = v
            if keys2:
                for k, v in zip(keys2, x[end:]):
                    variables[k] = v
    else:
        return
    del variables[key]


class Scan():

    def __new__(cls, *args, mixin=None, **kwds):
        if mixin is None:
            return super().__new__(cls)
        for k in dir(mixin):
            if not hasattr(cls, k):
                try:
                    setattr(cls, k, getattr(mixin, k))
                except:
                    pass
        return super().__new__(cls)

    def __init__(self,
                 app: str = 'task',
                 tags: tuple[str] = (),
                 database: str | Path
                 | None = default_server,
                 dump_globals: bool = False,
                 max_workers: int = 4,
                 max_promise: int = 100,
                 max_message: int = 1000,
                 config: dict | None = None,
                 mixin=None):
        self.id = task_uuid()
        self.record = None
        self.config = {} if config is None else copy.deepcopy(config)
        self._raw_config_copy = copy.deepcopy(self.config)
        self.description = {
            'app': app,
            'tags': tags,
            'config': None,
            'loops': {},
            'intrinsic_loops': {},
            'consts': {},
            'functions': {},
            'getters': {},
            'setters': {},
            'optimizers': {},
            'namespace': {} if dump_globals else None,
            'actions': {},
            'dependents': {},
            'order': {},
            'axis': {},
            'independent_variables': set(),
            'filters': {},
            'total': {},
            'database': database,
            'hiden': ['self', 'config', r'^__.*', r'.*__$', r'^#.*'],
            'entry': {
                'system': get_system_info(),
                'env': {},
                'shell': '',
                'cmds': [],
                'scripts': []
            },
        }
        self._current_level = 0
        self._variables = {}
        self._main_task = None
        self._background_tasks = ()
        self._sock = None
        self._sem = asyncio.Semaphore(max_promise + 1)
        self._bar: dict[int, tqdm] = {}
        self._hide_pattern_re = re.compile('|'.join(self.description['hiden']))
        self._msg_queue = asyncio.Queue(max_message)
        self._prm_queue = asyncio.Queue()
        self._single_step = True
        self._max_workers = max_workers
        self._max_promise = max_promise
        self._max_message = max_message
        self._executors = ProcessPoolExecutor(max_workers=max_workers)

    def __del__(self):
        try:
            self._main_task.cancel()
        except:
            pass

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state['record']
        del state['_sock']
        del state['_main_task']
        del state['_background_tasks']
        del state['_bar']
        del state['_msg_queue']
        del state['_prm_queue']
        del state['_sem']
        del state['_executors']
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self.record = None
        self._sock = None
        self._main_task = None
        self._background_tasks = ()
        self._bar = {}
        self._prm_queue = asyncio.Queue()
        self._msg_queue = asyncio.Queue(self._max_message)
        self._sem = asyncio.Semaphore(self._max_promise + 1)
        self._executors = ProcessPoolExecutor(max_workers=self._max_workers)
        for opt in self.description['optimizers'].values():
            opt.scanner = self

    def __del__(self):
        try:
            self._main_task.cancel()
        except:
            pass
        try:
            self._executors.shutdown()
        except:
            pass

    @property
    def current_level(self):
        return self._current_level

    @property
    def variables(self) -> dict[str, Any]:
        return self._variables

    async def _emit(self, current_level, step, position, variables: dict[str,
                                                                         Any]):
        for key, value in list(variables.items()):
            if key.startswith('*') or ',' in key:
                await _unpack(key, variables)
            elif inspect.isawaitable(value) and not self.hiden(key):
                variables[key] = await value

        if self.record is None:
            self.record = await self.create_record()
        if self._sock is not None:
            await self._sock.send_pyobj({
                'task': self.id,
                'method': 'record_append',
                'record_id': self.record.id,
                'level': current_level,
                'step': step,
                'position': position,
                'variables': {
                    k: v
                    for k, v in variables.items() if not self.hiden(k)
                }
            })
        else:
            self.record.append(current_level, step, position, {
                k: v
                for k, v in variables.items() if not self.hiden(k)
            })

    async def emit(self, current_level, step, position, variables: dict[str,
                                                                        Any]):
        await self._msg_queue.put(
            self._emit(current_level, step, position, variables.copy()))

    def hide(self, name: str):
        self.description['hiden'].append(name)
        self._hide_pattern_re = re.compile('|'.join(self.description['hiden']))

    def hiden(self, name: str) -> bool:
        return bool(self._hide_pattern_re.match(name)) or name.startswith(
            '*') or ',' in name

    async def _filter(self, variables: dict[str, Any], level: int = 0):
        try:
            return all(await asyncio.gather(*[
                call_function(fun, variables) for fun in itertools.chain(
                    self.description['filters'].get(level, []),
                    self.description['filters'].get(-1, []))
            ]))
        except:
            return True

    async def create_record(self):
        if self._sock is None:
            return Record(None, self.description['database'], self.description)

        if self.config:
            self.description['config'] = await create_config(
                self._raw_config_copy, self.description['database'],
                self._sock)
        if current_notebook() is None:
            await create_notebook('untitle', self.description['database'],
                                  self._sock)
        cell_id = await save_input_cells(current_notebook(),
                                         self.description['entry']['scripts'],
                                         self.description['database'],
                                         self._sock)
        self.description['entry']['scripts'] = cell_id

        await self._sock.send_pyobj({
            'task':
            self.id,
            'method':
            'record_create',
            'description':
            dump_dict(self.description,
                      keys=[
                          'intrinsic_loops', 'app', 'tags', 'loops',
                          'independent_variables', 'axis', 'config', 'entry'
                      ])
        })

        record_id = await self._sock.recv_pyobj()
        return Record(record_id, self.description['database'],
                      self.description)

    def get(self, name: str):
        if name in self.description['consts']:
            return self.description['consts'][name]
        else:
            try:
                return self._query_config(name)
            except:
                return Symbol(name)

    def _add_search_space(self, name: str, level: int, space):
        if level not in self.description['loops']:
            self.description['loops'][level] = []
        self.description['loops'][level].append((name, space))

    def add_depends(self, name: str, depends: list[str]):
        if isinstance(depends, str):
            depends = [depends]
        if 'self' in depends:
            depends.append('config')
        if name not in self.description['dependents']:
            self.description['dependents'][name] = set()
        self.description['dependents'][name].update(depends)

    def add_filter(self, func: Callable, level: int = -1):
        """
        Add a filter function to the scan.

        Args:
            func: A callable object or an instance of Expression.
            level: The level of the scan to add the filter. -1 means any level.
        """
        if level not in self.description['filters']:
            self.description['filters'][level] = []
        self.description['filters'][level].append(func)

    def set(self,
            name: str,
            value,
            depends: Iterable[str] | None = None,
            setter: Callable | None = None):
        try:
            dill.dumps(value)
        except:
            raise ValueError('value is not serializable.')
        if isinstance(value, Expression):
            self.add_depends(name, value.symbols())
            self.description['functions'][name] = value
        elif callable(value):
            if depends:
                self.add_depends(name, depends)
                s = ','.join(depends)
                self.description['functions'][f'#{name}'] = value
                self.description['functions'][name] = eval(
                    f"lambda self, {s}: self.description['functions']['#{name}']({s})"
                )
            else:
                self.add_depends(name, _get_depends(value))
                self.description['functions'][name] = value
        else:
            try:
                value = Space.fromarray(value)
            except:
                pass
            self.description['consts'][name] = value

        if '.' in name:
            self.add_depends('config', [name])

        if ',' in name:
            for key in name.split(','):
                if not key.startswith('*'):
                    self.add_depends(key, [name])
        if setter:
            self.description['setters'][name] = setter

    def search(self,
               name: str,
               space: Iterable | Expression | Callable | OptimizeSpace,
               level: int | None = None,
               setter: Callable | None = None,
               intrinsic: bool = False):
        if level is not None:
            if not intrinsic:
                assert level >= 0, 'level must be greater than or equal to 0.'
        if intrinsic:
            assert isinstance(space, (np.ndarray, list, tuple, range, Space)), \
                'space must be an instance of np.ndarray, list, tuple, range or Space.'
            self.description['intrinsic_loops'][name] = level
            self.set(name, space)
        elif isinstance(space, OptimizeSpace):
            space.name = name
            space.optimizer.dimensions[name] = space.space
            if space.suggestion is not None:
                space.optimizer.suggestion[name] = space.suggestion
            self._add_search_space(name, space.optimizer.level, space)
            self.add_depends(space.optimizer.name, [name])
        else:
            if level is None:
                raise ValueError('level must be provided.')
            try:
                space = Space.fromarray(space)
            except:
                pass
            self._add_search_space(name, level, space)
            if isinstance(space, Expression) or callable(space):
                self.add_depends(name, space.symbols())
        if setter:
            self.description['setters'][name] = setter
        if '.' in name:
            self.add_depends('config', [name])

    def trace(self,
              name: str,
              depends: list[str],
              getter: Callable | None = None):
        self.add_depends(name, depends)
        if getter:
            self.description['getters'][name] = getter

    def minimize(self,
                 name: str,
                 level: int,
                 method=NgOptimizer,
                 maxiter=100,
                 getter: Callable | None = None,
                 **kwds) -> Optimizer:
        assert level >= 0, 'level must be greater than or equal to 0.'
        opt = Optimizer(self,
                        name,
                        level,
                        method,
                        maxiter,
                        minimize=True,
                        **kwds)
        self.description['optimizers'][name] = opt
        if getter:
            self.description['getters'][name] = getter
        return opt

    def maximize(self,
                 name: str,
                 level: int,
                 method=NgOptimizer,
                 maxiter=100,
                 getter: Callable | None = None,
                 **kwds) -> Optimizer:
        assert level >= 0, 'level must be greater than or equal to 0.'
        opt = Optimizer(self,
                        name,
                        level,
                        method,
                        maxiter,
                        minimize=False,
                        **kwds)
        self.description['optimizers'][name] = opt
        if getter:
            self.description['getters'][name] = getter
        return opt

    def _synchronize_config(self):
        for key, value in self.variables.items():
            if '.' in key:
                d = self.config
                ks = key.split('.')
                if not ks:
                    continue
                for k in ks[:-1]:
                    if k in d:
                        d = d[k]
                    else:
                        d[k] = {}
                        d = d[k]
                d[ks[-1]] = value
        return self.config

    def _query_config(self, key):
        d = self.config
        for k in key.split('.'):
            d = d[k]
        return d

    async def _update_progress(self):
        while True:
            task = await self._prm_queue.get()
            await task
            self._prm_queue.task_done()

    async def _send_msg(self):
        while True:
            task = await self._msg_queue.get()
            await task
            self._msg_queue.task_done()

    @contextlib.asynccontextmanager
    async def _send_msg_and_update_bar(self):
        send_msg_task = asyncio.create_task(self._send_msg())
        update_progress_task = asyncio.create_task(self._update_progress())
        try:
            yield (send_msg_task, update_progress_task)
        finally:
            update_progress_task.cancel()
            send_msg_task.cancel()
            while True:
                try:
                    task = self._prm_queue.get_nowait()
                except:
                    break
                try:
                    task.cancel()
                except:
                    pass

    async def _check_background_tasks(self):
        for task in self._background_tasks:
            if task.done():
                await task

    async def run(self):
        assymbly(self.description)
        self._background_tasks = ()

        if isinstance(
                self.description['database'],
                str) and self.description['database'].startswith("tcp://"):
            async with ZMQContextManager(zmq.DEALER,
                                         connect=self.description['database'],
                                         socket=self._sock) as socket:
                self._sock = socket
                async with self._send_msg_and_update_bar() as background_tasks:
                    self._background_tasks = background_tasks
                    await self._run()
        else:
            if self.config:
                self.description['config'] = self._raw_config_copy
            async with self._send_msg_and_update_bar() as background_tasks:
                self._background_tasks = background_tasks
                await self._run()

    async def _run(self):
        self._variables = {'self': self, 'config': self.config}

        consts = {}
        for k, v in self.description['consts'].items():
            if isinstance(v, Space):
                consts[k] = v.toarray()
            else:
                consts[k] = v

        await update_variables(self._variables, consts,
                               self.description['setters'])
        for level, total in self.description['total'].items():
            if total == np.inf:
                total = None
            self._bar[level] = tqdm(total=total)

        updates = await call_many_functions(
            self.description['order'].get(-1, []),
            self.description['functions'], self.variables)
        await update_variables(self.variables, updates,
                               self.description['setters'])
        await self._check_background_tasks()
        await self.work()
        for level, bar in self._bar.items():
            bar.close()
        await self._check_background_tasks()
        if self._single_step:
            self.variables.update(await call_many_functions(
                self.description['order'].get(-1, []),
                self.description['getters'], self.variables))

            await self.emit(0, 0, 0, self.variables)
            await self.emit(-1, 0, 0, {})
        await self._check_background_tasks()
        await self._prm_queue.join()
        await self._msg_queue.join()
        return self.variables

    async def done(self):
        if self._main_task is not None:
            try:
                await self._main_task
            except asyncio.CancelledError:
                pass

    def finished(self):
        return self._main_task.done()

    def start(self):
        import asyncio
        self._main_task = asyncio.create_task(self.run())

    async def submit(self, server=default_executor):
        assymbly(self.description)
        async with ZMQContextManager(zmq.DEALER,
                                     connect=server,
                                     socket=self._sock) as socket:
            await socket.send_pyobj({
                'method': 'task_submit',
                'description': dill.dumps(self.description)
            })
            self.id = await socket.recv_pyobj()
            await socket.send_pyobj({
                'method': 'task_get_record_id',
                'id': self.id
            })
            record_id = await socket.recv_pyobj()
            self.record = Record(record_id, self.description['database'],
                                 self.description)

    def cancel(self):
        if self._main_task is not None:
            self._main_task.cancel()

    async def _reset_progress_bar(self, level):
        if level in self._bar:
            self._bar[level].reset()

    async def _update_progress_bar(self, level, n: int):
        if level in self._bar:
            self._bar[level].update(n)

    async def iter(self, **kwds):
        if self.current_level >= len(self.description['loops']):
            return
        step = 0
        position = 0
        self._prm_queue.put_nowait(self._reset_progress_bar(
            self.current_level))
        async for variables in _iter_level(
                self.variables,
                self.description['loops'].get(self.current_level, []),
                self.description['order'].get(self.current_level, []),
                self.description['functions']
                | {'config': self._synchronize_config},
                self.description['optimizers'], self.description['setters'],
                self.description['getters']):
            await self._check_background_tasks()
            self._current_level += 1
            if await self._filter(variables, self.current_level - 1):
                yield variables
                self._single_step = False
                await self.emit(self.current_level - 1, step, position,
                                variables)
                step += 1
            position += 1
            self._current_level -= 1
            self._prm_queue.put_nowait(
                self._update_progress_bar(self.current_level, 1))
            await self._check_background_tasks()
        if self.current_level == 0:
            await self.emit(self.current_level - 1, 0, 0, {})
            for name, value in self.variables.items():
                if inspect.isawaitable(value):
                    self.variables[name] = await value
            await self._check_background_tasks()
            await self._prm_queue.join()

    async def work(self, **kwds):
        if self.current_level in self.description['actions']:
            action = self.description['actions'][self.current_level]
            coro = action(self, **kwds)
            if inspect.isawaitable(coro):
                await coro
        else:
            async for variables in self.iter(**kwds):
                await self.do_something(**kwds)

    async def do_something(self, **kwds):
        await self.work(**kwds)

    def mount(self, action: Callable, level: int):
        """
        Mount a action to the scan.

        Args:
            action: A callable object.
            level: The level of the scan to mount the action.
        """
        self.description['actions'][level] = action

    async def promise(self, awaitable: Awaitable | Callable, *args,
                      **kwds) -> Promise:
        """
        Promise to calculate asynchronous function and return the result in future.

        Args:
            awaitable: An awaitable object.

        Returns:
            Promise: A promise object.
        """
        if inspect.isawaitable(awaitable):
            async with self._sem:
                task = asyncio.create_task(self._await(awaitable))
                self._prm_queue.put_nowait(task)
                return Promise(task)
        elif inspect.iscoroutinefunction(awaitable):
            return await self.promise(awaitable(*args, **kwds))
        elif callable(awaitable):
            try:
                buf = dill.dumps((awaitable, args, kwds))
                task = asyncio.get_running_loop().run_in_executor(
                    self._executors, _run_function_in_process, buf)
            except:
                return awaitable(*args, **kwds)
            self._prm_queue.put_nowait(task)
            return Promise(task)
        else:
            return awaitable

    async def _await(self, awaitable: Awaitable):
        async with self._sem:
            return await awaitable


def _get_environment(description):
    import __main__
    from IPython import get_ipython

    if isinstance(description['namespace'], dict):
        description['namespace'] = dump_globals()

    ipy = get_ipython()
    if ipy is not None:
        description['entry']['shell'] = 'ipython'
        description['entry']['scripts'] = [
            yapf_reformat(cell_text) for cell_text in ipy.user_ns['In']
        ]
    else:
        try:
            description['entry']['shell'] = 'shell'
            description['entry']['cmds'] = [
                sys.executable, __main__.__file__, *sys.argv[1:]
            ]
            description['entry']['scripts'] = []
            try:
                with open(__main__.__file__) as f:
                    description['entry']['scripts'].append(f.read())
            except:
                pass
        except:
            pass

    description['entry']['env'] = {k: v for k, v in os.environ.items()}

    return description


def _mapping_levels(description):
    mapping = {
        label: level
        for level, label in enumerate(
            sorted(
                set(description['loops'].keys())
                | {k
                   for k in description['actions'].keys() if k >= 0}))
    }

    if -1 in description['actions']:
        mapping[-1] = max(mapping.values()) + 1

    levels = sorted(mapping.values())
    for k in description['actions'].keys():
        if k < -1:
            mapping[k] = levels[k]

    description['loops'] = dict(
        sorted([(mapping[k], v) for k, v in description['loops'].items()]))
    description['actions'] = {
        mapping[k]: v
        for k, v in description['actions'].items()
    }

    for level, loops in description['loops'].items():
        description['total'][level] = np.inf
        for name, space in loops:
            try:
                description['total'][level] = min(description['total'][level],
                                                  len(space))
            except:
                pass
    return levels


def _get_independent_variables(description):
    independent_variables = set(description['intrinsic_loops'].keys())
    for level, loops in description['loops'].items():
        for name, iterable in loops:
            if isinstance(iterable, (np.ndarray, list, tuple, range, Space)):
                independent_variables.add(name)
    return independent_variables


def _build_dependents(description, levels, independent_variables):
    dependents = copy.deepcopy(description['dependents'])
    all_nodes = set(description['dependents'].keys())
    for key, deps in dependents.items():
        all_nodes.update(deps)

    for level in levels:
        range_list = description['loops'].get(level, [])
        if level > 0:
            if f'#__loop_{level}' not in description['dependents']:
                dependents[f'#__loop_{level}'] = set()
            dependents[f'#__loop_{level}'].add(f'#__loop_{level-1}')
        for name, _ in range_list:
            if name not in description['dependents']:
                dependents[name] = set()
            dependents[name].add(f'#__loop_{level}')

    after_yield = set()
    for key in all_nodes:
        if key not in independent_variables and key not in description[
                'consts']:
            if key not in dependents:
                after_yield.add(key)

    def _get_all_depends(key, graph):
        ret = set()
        if key not in graph:
            return ret

        for e in graph[key]:
            ret.update(_get_all_depends(e, graph))
        ret.update(graph[key])
        return ret

    full_depends = {}
    for key in dependents:
        full_depends[key] = _get_all_depends(key, dependents)
        if full_depends[key] & after_yield:
            after_yield.add(key)

    return dependents, full_depends, after_yield


def _build_order(description, levels, dependents, full_depends):
    levels = {}
    passed = set()
    all_keys = set(description['consts'].keys())
    for key in dependents:
        all_keys.add(key)
        all_keys.update(dependents[key])
    for level in reversed(description['loops'].keys()):
        tag = f'#__loop_{level}'
        for key, deps in full_depends.items():
            all_keys.update(deps)
            all_keys.add(key)
            if key.startswith('#__loop_'):
                continue
            if tag in deps:
                if level not in levels:
                    levels[level] = set()
                if key not in passed:
                    passed.add(key)
                    levels[level].add(key)
    levels[-1] = {
        key
        for key in all_keys - passed if not key.startswith('#__loop_')
    }

    order = []
    ts = TopologicalSorter(dependents)
    ts.prepare()
    while ts.is_active():
        ready = ts.get_ready()
        order.append(ready)
        for k in ready:
            ts.done(k)

    description['order'] = {}

    for level in sorted(levels):
        keys = set(levels[level])
        description['order'][level] = []
        for ready in order:
            ready = list(keys & set(ready))
            if ready:
                description['order'][level].append(ready)
                keys -= set(ready)


def _make_axis(description):
    axis = {}

    for name in description['consts']:
        axis[name] = ()
    for level, range_list in description['loops'].items():
        for name, iterable in range_list:
            if isinstance(iterable, OptimizeSpace):
                axis[name] = tuple(range(level + 1))
                continue
            axis[name] = (level, )

    for level, group in description['order'].items():
        for names in group:
            for name in names:
                if name not in description['dependents']:
                    if name not in axis:
                        axis[name] = (level, )
                else:
                    d = set()
                    for n in description['dependents'][name]:
                        d.update(axis[n])
                    if name not in axis:
                        axis[name] = tuple(sorted(d))
                    else:
                        axis[name] = tuple(sorted(set(axis[name]) | d))
    description['axis'] = {
        k: tuple([x for x in v if x >= 0])
        for k, v in axis.items()
    }


def assymbly(description):
    _get_environment(description)
    levels = _mapping_levels(description)
    independent_variables = _get_independent_variables(description)
    description['independent_variables'] = independent_variables

    dependents, full_depends, after_yield = _build_dependents(
        description, levels, independent_variables)

    _build_order(description, levels, dependents, full_depends)
    _make_axis(description)

    return description



import hashlib
import lzma
import pickle
import re
import uuid
import zipfile
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from functools import lru_cache
from pathlib import Path
from typing import Any, Literal
from urllib.parse import parse_qs

from loguru import logger

try:
    from paramiko import SSHClient
    from paramiko.ssh_exception import SSHException
except:
    import warnings

    warnings.warn("Can't import paramiko, ssh support will be disabled.")

    class SSHClient:

        def __init__(self):
            raise ImportError(
                "Can't import paramiko, ssh support will be disabled.")

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            pass

    class SSHException(Exception):
        pass


from ..cli.config import get_config_value

__current_config_cache = None


@dataclass
class Report():
    workflow: str = ''
    in_spec: bool = False
    bad_data: bool = False
    fully_calibrated: bool = False
    calibrated_time: datetime = field(default_factory=datetime.now)
    checked_time: datetime = field(default_factory=datetime.now)
    ttl: timedelta = timedelta(days=3650)
    parameters: dict = field(default_factory=dict)
    oracle: dict = field(default_factory=dict)
    other_infomation: dict = field(default_factory=dict)
    data: Any = field(default_factory=tuple, repr=False)
    index: int = -1
    previous_path: Path | None = field(default=None, repr=False)
    heads: dict[str, Path] = field(default_factory=dict, repr=False)
    base_path: str | Path | None = field(default=None, repr=False)
    path: Path | None = field(default=None, repr=False)
    config_path: Path | None = field(default=None, repr=False)
    script_path: Path | None = field(default=None, repr=False)

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop('base_path', None)
        for k in ['path', 'previous_path', 'config_path', 'script_path']:
            if state[k] is not None:
                state[k] = str(state[k])
        return state

    def __setstate__(self, state):
        for k in ['path', 'previous_path', 'config_path', 'script_path']:
            if state[k] is not None:
                state[k] = Path(state[k])
        self.__dict__.update(state)

    @property
    def previous(self):
        if self.previous_path is not None and self.base_path is not None:
            return load_report(self.previous_path, self.base_path)
        else:
            return None

    @property
    def state(self) -> Literal['OK', 'Bad', 'Outdated']:
        state = 'Bad'
        match (self.in_spec, self.bad_data):
            case (True, False):
                state = 'OK'
            case (False, True):
                state = 'Bad'
            case (False, False):
                state = 'Outdated'
        return state

    @state.setter
    def state(self, state: Literal['OK', 'Bad', 'Outdated', 'In spec',
                                   'Out of spec', 'Bad data']):
        if state not in [
                'OK', 'Bad', 'Outdated', 'In spec', 'Out of spec', 'Bad data'
        ]:
            raise ValueError(
                f'Invalid state: {state}, state must be one of "OK", "Bad" and "Outdated"'
            )
        if state in ['In spec', 'OK']:
            self.in_spec = True
            self.bad_data = False
        elif state in ['Bad data', 'Bad']:
            self.bad_data = True
            self.in_spec = False
        else:
            self.bad_data = False
            self.in_spec = False

    @property
    def config(self):
        if self.config_path is not None and self.base_path is not None:
            return load_item(self.config_path, 'items', self.base_path)
        else:
            return None

    @property
    def script(self):
        if self.script_path is not None and self.base_path is not None:
            source = load_item(self.script_path, 'items', self.base_path)
            if isinstance(source, str):
                return source
            else:
                from .template import inject_mapping
                return inject_mapping(*source)[0]
        else:
            return None

    @property
    def template_source(self):
        if self.script_path is not None and self.base_path is not None:
            source = load_item(self.script_path, 'items', self.base_path)
            return source
        else:
            return None


def random_path(base: Path) -> Path:
    while True:
        s = uuid.uuid4().hex
        path = Path(s[:2]) / s[2:4] / s[4:6] / s[6:]
        if not (base / path).exists():
            return path


def find_report(
    workflow: str, base_path: str | Path = get_config_value("data", Path)
) -> Report | None:
    if workflow.startswith("cfg:"):
        return find_config_key_history(workflow[4:], base_path)

    base_path = Path(base_path)
    path = get_head(workflow, base_path)
    if path is None:
        return None
    return load_report(path, base_path)


def renew_report(workflow: str, report: Report | None, base_path: str | Path):
    logger.debug(f'Renewing report for "{workflow}"')
    if report is not None:
        report.checked_time = datetime.now()
        return save_report(workflow,
                           report,
                           base_path,
                           overwrite=True,
                           refresh_heads=True)
    else:
        raise ValueError(f"Can't renew report for {workflow}")


def revoke_report(workflow: str, report: Report | None, base_path: str | Path):
    logger.debug(f'Revoking report for "{workflow}"')
    base_path = Path(base_path)
    if report is not None:
        report.in_spec = False
        report.previous_path = report.path
        return save_report(workflow,
                           report,
                           base_path,
                           overwrite=False,
                           refresh_heads=True)


def get_report_by_index(
    index: int, base_path: str | Path = get_config_value("data", Path)
) -> Report | None:
    try:
        path = query_index("report", base_path, index)
        return load_report(path, base_path)
    except:
        raise
        return None


def get_head(workflow: str, base_path: str | Path) -> Path | None:
    return get_heads(base_path).get(workflow, None)


#########################################################################
##                           Basic Write API                           ##
#########################################################################


def set_head(workflow: str, path: Path, base_path: str | Path):
    base_path = Path(base_path)
    base_path.mkdir(parents=True, exist_ok=True)
    try:
        with open(base_path / "heads", "rb") as f:
            heads = pickle.load(f)
    except:
        heads = {}
    heads[workflow] = path
    with open(base_path / "heads", "wb") as f:
        pickle.dump(heads, f)


def save_report(workflow: str,
                report: Report,
                base_path: str | Path,
                overwrite: bool = False,
                refresh_heads: bool = True) -> int:
    if workflow.startswith("cfg:"):
        return save_config_key_history(workflow[4:], report, base_path)

    logger.debug(
        f'Saving report for "{workflow}", {report.in_spec=}, {report.bad_data=}, {report.fully_calibrated=}'
    )
    base_path = Path(base_path)
    try:
        buf = lzma.compress(pickle.dumps(report))
    except:
        raise ValueError(f"Can't pickle report for {workflow}")
    if overwrite:
        path = report.path
        if path is None:
            raise ValueError("Report path is None, can't overwrite.")
        with open(base_path / 'reports' / path, "rb") as f:
            index = int.from_bytes(f.read(8), 'big')
        report.index = index
    else:
        path = random_path(base_path / 'reports')
        (base_path / 'reports' / path).parent.mkdir(parents=True,
                                                    exist_ok=True)
        report.path = path
        report.index = create_index("report",
                                    base_path,
                                    context=str(path),
                                    width=35)
    with open(base_path / 'reports' / path, "wb") as f:
        f.write(report.index.to_bytes(8, 'big'))
        f.write(buf)
    if refresh_heads:
        set_head(workflow, path, base_path)
    return report.index


def create_index(name: str,
                 base_path: str | Path,
                 context: str,
                 width: int = -1,
                 start: int = 0):

    path = Path(base_path) / "index" / name
    if width == -1:
        width = len(context)
    else:
        width = max(width, len(context))

    if path.with_suffix('.width').exists():
        width = int(path.with_suffix('.width').read_text())
        index = int(path.with_suffix('.seq').read_text())
    else:
        index = start
    if width < len(context):
        raise ValueError(
            f"Context '{context}' is too long, existing width of '{name}' is {width}."
        )
    if not path.with_suffix('.width').exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        path.with_suffix('.width').write_text(str(width))

    path.with_suffix('.seq').write_text(str(index + 1))

    with path.with_suffix('.idx').open("a") as f:
        f.write(f"{context.ljust(width)}\n")

    return index


def save_item(item, group, data_path):
    salt = 0
    buf = pickle.dumps(item)
    buf = lzma.compress(buf)
    h = hashlib.md5(buf)

    while True:
        h.update(f"{salt}".encode())
        hashstr = h.hexdigest()
        item_id = Path(hashstr[:2]) / hashstr[2:4] / hashstr[4:]
        path = Path(data_path) / group / item_id
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            with open(path, 'wb') as f:
                f.write(buf)
            break
        elif path.read_bytes() == buf:
            break
        salt += 1
    return str(item_id)


def append_item_data(data, id, group, base_path):
    path = Path(base_path) / group / id
    if not path.exists():
        raise ValueError(f"Item {id} does not exist.")
    with open(path, 'ab') as f:
        f.write(data)


def save_config_key_history(key: str, report: Report,
                            base_path: str | Path) -> int:
    global __current_config_cache
    base_path = Path(base_path) / 'state'
    base_path.mkdir(parents=True, exist_ok=True)

    if __current_config_cache is None:
        if (base_path / 'parameters.pkl').exists():
            with open(base_path / 'parameters.pkl', 'rb') as f:
                __current_config_cache = pickle.load(f)
        else:
            __current_config_cache = {}

    __current_config_cache[
        key] = report.data, report.calibrated_time, report.checked_time

    with open(base_path / 'parameters.pkl', 'wb') as f:
        pickle.dump(__current_config_cache, f)
    return 0


#########################################################################
##                            Basic Read API                           ##
#########################################################################


def load_report(path: str | Path, base_path: str | Path) -> Report | None:
    if isinstance(base_path, str) and base_path.startswith('ssh://'):
        with SSHClient() as client:
            cfg = parse_ssh_uri(base_path)
            remote_base_path = cfg.pop('remote_file_path')
            client.load_system_host_keys()
            client.connect(**cfg)
            report = load_report_from_scp(path, remote_base_path, client)
            report.base_path = base_path
            return report

    base_path = Path(base_path)
    if zipfile.is_zipfile(base_path):
        return load_report_from_zipfile(path, base_path)

    path = base_path / 'reports' / path

    with open(base_path / 'reports' / path, "rb") as f:
        index = int.from_bytes(f.read(8), 'big')
        report = pickle.loads(lzma.decompress(f.read()))
        report.base_path = base_path
        report.index = index
        return report


def get_heads(base_path: str | Path) -> Path | None:
    if isinstance(base_path, str) and base_path.startswith('ssh://'):
        with SSHClient() as client:
            cfg = parse_ssh_uri(base_path)
            remote_base_path = cfg.pop('remote_file_path')
            client.load_system_host_keys()
            client.connect(**cfg)
            return get_heads_from_scp(remote_base_path, client)

    base_path = Path(base_path)
    if zipfile.is_zipfile(base_path):
        return get_heads_from_zipfile(base_path)
    try:
        with open(base_path / "heads", "rb") as f:
            heads = pickle.load(f)
        return heads
    except:
        return {}


@lru_cache(maxsize=4096)
def query_index(name: str, base_path: str | Path, index: int):
    if isinstance(base_path, str) and base_path.startswith('ssh://'):
        with SSHClient() as client:
            cfg = parse_ssh_uri(base_path)
            remote_base_path = cfg.pop('remote_file_path')
            client.load_system_host_keys()
            client.connect(**cfg)
            return query_index_from_scp(name, remote_base_path, client, index)

    base_path = Path(base_path)
    if zipfile.is_zipfile(base_path):
        return query_index_from_zipfile(name, base_path, index)
    path = Path(base_path) / "index" / name
    width = int(path.with_suffix('.width').read_text())

    with path.with_suffix('.idx').open("r") as f:
        f.seek(index * (width + 1))
        context = f.read(width)
    return context.rstrip()


@lru_cache(maxsize=4096)
def load_item(id, group, base_path):
    if isinstance(base_path, str) and base_path.startswith('ssh://'):
        with SSHClient() as client:
            cfg = parse_ssh_uri(base_path)
            remote_base_path = cfg.pop('remote_file_path')
            client.load_system_host_keys()
            client.connect(**cfg)
            buf = load_item_buf_from_scp(id, group, remote_base_path, client)
    else:
        base_path = Path(base_path)
        if zipfile.is_zipfile(base_path):
            buf = load_item_buf_from_zipfile(id, group, base_path)
        else:
            path = Path(base_path) / group / id
            with open(path, 'rb') as f:
                buf = f.read()
    item = pickle.loads(lzma.decompress(buf))
    return item


def find_config_key_history(key: str, base_path: str | Path) -> Report | None:
    global __current_config_cache
    base_path = Path(base_path) / 'state'
    if __current_config_cache is None:
        if (base_path / 'parameters.pkl').exists():
            with open(base_path / 'parameters.pkl', 'rb') as f:
                __current_config_cache = pickle.load(f)
        else:
            __current_config_cache = {}

    if key in __current_config_cache:
        value, calibrated_time, checked_time = __current_config_cache.get(
            key, None)
        report = Report(
            workflow=f'cfg:{key}',
            bad_data=False,
            in_spec=True,
            fully_calibrated=True,
            parameters={key: value},
            data=value,
            calibrated_time=calibrated_time,
            checked_time=checked_time,
        )
        return report
    return None


#########################################################################
##                            Zipfile support                          ##
#########################################################################


def load_report_from_zipfile(path: str | Path,
                             base_path: str | Path) -> Report | None:
    path = Path(path)
    with zipfile.ZipFile(base_path) as zf:
        path = '/'.join(path.parts)
        with zf.open(f"{base_path.stem}/reports/{path}") as f:
            index = int.from_bytes(f.read(8), 'big')
            report = pickle.loads(lzma.decompress(f.read()))
            report.base_path = base_path
            report.index = index
            return report


def get_heads_from_zipfile(base_path: str | Path) -> Path | None:
    with zipfile.ZipFile(base_path) as zf:
        with zf.open(f"{base_path.stem}/heads") as f:
            heads = pickle.load(f)
    return heads


def query_index_from_zipfile(name: str, base_path: str | Path, index: int):
    with zipfile.ZipFile(base_path) as zf:
        with zf.open(f"{base_path.stem}/index/{name}.width") as f:
            width = int(f.read().decode())
        with zf.open(f"{base_path.stem}/index/{name}.idx") as f:
            f.seek(index * (width + 1))
            context = f.read(width).decode()
    return context.rstrip()


def load_item_buf_from_zipfile(id, group, base_path):
    with zipfile.ZipFile(base_path) as zf:
        with zf.open(f"{base_path.stem}/{group}/{id}") as f:
            return f.read()


#########################################################################
##                             SCP support                             ##
#########################################################################


def parse_ssh_uri(uri):
    """
    解析 SSH URI 字符串，返回包含连接参数和路径的字典。
    
    格式：ssh://[{username}[:{password}]@]{host}[:{port}][?key_filename={key_path}][/{remote_file_path}]

    返回示例：
    {
        "username": "user",
        "password": "pass",
        "host": "example.com",
        "port": 22,
        "key_filename": "/path/to/key",
        "remote_file_path": "/data/file.txt"
    }
    """
    pattern = re.compile(
        r"^ssh://"  # 协议头
        r"(?:([^:@/]+))(?::([^@/]+))?@?"  # 用户名和密码（可选）
        r"([^:/?#]+)"  # 主机名（必须）
        r"(?::(\d+))?"  # 端口（可选）
        r"(/?[^?#]*)?"  # 远程路径（可选）
        r"(?:\?([^#]+))?"  # 查询参数（如 key_filename）
        r"$",
        re.IGNORECASE)

    match = pattern.match(uri)
    if not match:
        raise ValueError(f"Invalid SSH URI format: {uri}")

    # 提取分组
    username, password, host, port, path, query = match.groups()

    # 处理查询参数
    key_filename = None
    if query:
        params = parse_qs(query)
        key_filename = params.get("key_filename", [None])[0]  # 取第一个值

    # 清理路径开头的斜杠
    remote_file_path = path

    return {
        "username": username,
        "password": password,
        "hostname": host,
        "port": int(port) if port else 22,  # 默认端口 22
        "key_filename": key_filename,
        "remote_file_path": remote_file_path
    }


def load_report_from_scp(path: str | Path, base_path: Path,
                         client: SSHClient) -> Report:
    try:
        path = Path(path)
        with client.open_sftp() as sftp:
            with sftp.open(str(Path(base_path) / 'reports' / path), 'rb') as f:
                index = int.from_bytes(f.read(8), 'big')
                report = pickle.loads(lzma.decompress(f.read()))
                report.index = index
                return report
    except SSHException:
        raise ValueError(f"Can't load report from {path}")


def get_heads_from_scp(base_path: Path, client: SSHClient) -> Path | None:
    try:
        with client.open_sftp() as sftp:
            with sftp.open(str(Path(base_path) / 'heads'), 'rb') as f:
                heads = pickle.load(f)
        return heads
    except SSHException:
        return None


def query_index_from_scp(name: str, base_path: Path, client: SSHClient,
                         index: int):
    try:
        with client.open_sftp() as sftp:
            s = str(Path(base_path) / 'index' / f'{name}.width')
            with sftp.open(s, 'rb') as f:
                width = int(f.read().decode())
            with sftp.open(str(Path(base_path) / 'index' / f'{name}.idx'),
                           'rb') as f:
                f.seek(index * (width + 1))
                context = f.read(width).decode()
        return context.rstrip()
    except SSHException:
        return None


def load_item_buf_from_scp(id: str, group: str, base_path: Path,
                           client: SSHClient):
    try:
        with client.open_sftp() as sftp:
            with sftp.open(str(Path(base_path) / group / str(id)), 'rb') as f:
                return f.read()
    except SSHException:
        return None



import bisect
import pathlib
from itertools import chain

import dill

from .base_dataset import BaseDataset
from .file import ArrayFile, BinaryFile, ObjectFile, ObjectListFile, load


class Dataset(BaseDataset):

    def __init__(self, uuid, storage):
        super().__init__()
        self.uuid = uuid
        self.storage = storage
        try:
            obj = self.storage.get(self.uuid)
            self.data = obj.data
            self.pos = obj.pos
            self.timestamps = obj.timestamps
            self.iteration = obj.iteration
            self.depends = obj.depends
            self.shape = obj.shape
            self.dims = obj.dims
            self.vars_dims = obj.vars_dims
            self.ctime = obj.ctime
            self.mtime = obj.mtime
            self._init_keys = obj._init_keys
            self._frozen_keys = obj._frozen_keys
            self._key_levels = obj._key_levels
            self.save_kwds = obj.save_kwds
        except:
            pass

    def unlink(self):
        for file in pathlib.Path(self._path).iterdir():
            if file.name.startswith('storage_'):
                file.unlink()
            elif file.name.startswith('pos_'):
                file.unlink()
            elif file.name.startswith('timestamp_'):
                file.unlink()
            elif file.name.startswith('iteration_'):
                file.unlink()
            elif file.name == 'index':
                file.unlink()
        try:
            pathlib.Path(self._path).unlink()
        except:
            pass

    def _append(self, iteration, pos, dataframe, kwds, now):
        for k, v in chain(kwds.items(), dataframe.items()):
            if k in self._frozen_keys:
                continue
            if k.startswith('__'):
                continue
            if self.vars_dims.get(k, ()) == () and k not in dataframe:
                continue
            self.count += 1
            if k not in self.data:
                self.data[k] = ArrayFile(self._path / f'storage_{k}', v)
                if k in self.vars_dims:
                    self.pos[k] = tuple(
                        ArrayFile(self._path / f'pos_{k}_{j}', pos[i])
                        for j, i in enumerate(self.vars_dims[k]))
                else:
                    self.pos[k] = tuple(
                        ArrayFile(self._path / f'pos_{k}_{j}', i)
                        for j, i in enumerate(pos))
                self.timestamps[k] = ArrayFile(self._path / f'timestamp_{k}',
                                               now.timestamp())
                self.iteration[k] = ArrayFile(self._path / f'iteration_{k}',
                                              iteration)
            else:
                if k in self.vars_dims:
                    pos_k = tuple(pos[i] for i in self.vars_dims[k])
                    if k not in dataframe and pos_k in zip(
                            *[list(l) for l in self.pos[k]]):
                        continue
                    for i, l in zip(pos_k, self.pos[k]):
                        l.append(i)
                else:
                    for i, l in zip(pos, self.pos[k]):
                        l.append(i)
                self.timestamps[k].append(now.timestamp())
                self.iteration[k].append(iteration)
                self.data[k].append(v)
        with open(self._path / 'index', 'wb') as f:
            dill.dump(self, f)

    def _get_array(self, key, shape, count):
        import numpy as np

        if key in self.vars_dims:
            shape = tuple([shape[i] for i in self.vars_dims[key]])

        data, data_shape, data_count = self.cache.get(key, (None, (), 0))
        if (data_shape, data_count) == (shape, count):
            return data
        try:
            tmp = np.asarray(list(self.data[key]))
            if data_shape != shape:
                data = np.full(shape + tmp.shape[1:], np.nan, dtype=tmp.dtype)
        except:
            tmp = list(self.data[key])
            if data_shape != shape:
                data = np.full(shape, np.nan, dtype=object)
        try:
            pos = tuple([list(l) for l in self.pos[key]])
            data[pos] = tmp
        except:
            print(key)
            print(data)
            print(pos)
            print(tmp)
            raise
        self.cache[key] = (data, shape, count)
        return data

    def _get_part(self, key, skip):
        i = bisect.bisect_left(list(self.iteration[key]), skip)
        pos = tuple(list(p)[i:] for p in self.pos[key])
        iteration = list(self.iteration[key])[i:]
        data = list(self.data[key])[i:]
        return data, iteration, pos



"""
This module provides the visualization of the qdat file.
"""
import logging
import math
import re
from functools import reduce

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm

log = logging.getLogger(__name__)


def draw(record_dict, fig=None, transpose=True, remove=True):
    """Draw a 2D or 3D plot from a record dict.

    Args:
        record_dict (dict): The record dict.
        fig (matplotlib.figure.Figure): The figure to draw on.
        transpose (bool): Whether to transpose the data.
    """
    dim = record_dict['info']['dim']
    zshape = record_dict['info']['zshape']
    zsize = reduce(lambda a, b: a * b, zshape, 1)

    axis_label = np.asarray(record_dict['info'].get('axis_label',
                                                    [])).flatten()
    z_label = np.asarray(record_dict['info'].get('z_label', [])).flatten()
    axis_unit = np.asarray(record_dict['info'].get('axis_unit', [])).flatten()
    z_unit = np.asarray(record_dict['info'].get('z_unit', [])).flatten()

    fig = plt.figure() if fig is None else fig
    if zsize < 4:
        plot_shape = (1, zsize)
    else:
        n = int(np.sqrt(zsize))
        plot_shape = (math.ceil(zsize / n), n)
    if dim < 3 and zsize < 100:
        figsize = plot_shape[1] * 8, plot_shape[0] * 6
        fig.set_size_inches(*figsize)
        axes = fig.subplots(*plot_shape)
        axes = np.array(axes).flatten()
    else:
        axes = []

    cb_list = []
    if dim == 1 and zsize < 101:
        for i in range(zsize):
            title = record_dict[
                'name'] + f' {i}' if zsize > 1 else record_dict['name']
            try:
                xlabel = axis_label[0]
                ylabel = z_label[0] if len(z_label) == 1 else z_label[i]
            except:
                xlabel, ylabel = 'X', 'Y'
            try:
                x_unit = axis_unit[0]
                y_unit = z_unit[0] if len(z_unit) == 1 else z_unit[i]
            except:
                x_unit, y_unit = 'a.u.', 'a.u.'
            axes[i].set_title(title)
            axes[i].set_xlabel(f'{xlabel} ({x_unit})')
            axes[i].set_ylabel(f'{ylabel} ({y_unit})')
    elif dim == 2 and zsize < 101:
        for i in range(zsize):
            smp = cm.ScalarMappable(norm=None, cmap=None)
            cb = fig.colorbar(smp, ax=axes[i])  # 默认色谱的colorbar
            cb_list.append(cb)
            try:
                title = record_dict['name'] + \
                    f': {z_label[i]}' if zsize > 1 else record_dict['name']
            except:
                title = record_dict[
                    'name'] + f' {i}' if zsize > 1 else record_dict['name']
            try:
                xlabel, ylabel = axis_label[1], axis_label[0]
                if transpose:
                    xlabel, ylabel = ylabel, xlabel
            except:
                xlabel, ylabel = 'X', 'Y'
            try:
                x_unit, y_unit = axis_unit[1], axis_unit[0]
                if transpose:
                    x_unit, y_unit = y_unit, x_unit
            except:
                x_unit, y_unit = 'a.u.', 'a.u.'
            axes[i].set_title(title)
            axes[i].set_xlabel(f'{xlabel} ({x_unit})')
            axes[i].set_ylabel(f'{ylabel} ({y_unit})')
    else:
        message1 = f'dim {dim} is too large (>2)! ' if dim > 2 else f'dim={dim}; '
        message2 = f'zsize {zsize} is too large (>101)!' if zsize > 101 else f'zsize={zsize}'
        log.warning('PASS: ' + message1 + message2)
    #########################################################################################
    try:
        tags = [
            tag.strip(r'\*')
            for tag in record_dict['ParaSpace'].get('tags', [])
            if re.match(r'\*', tag)
        ]
        tag_text = ','.join(tags)
        if tag_text:
            axes[0].text(-0.1,
                         1.1,
                         'TAG: ' + tag_text,
                         horizontalalignment='left',
                         verticalalignment='bottom',
                         transform=axes[0].transAxes)  # fig.transFigure)#
    except:
        pass
    #########################################################################################
    fig.tight_layout()

    dim = record_dict['info']['dim']
    zshape = record_dict['info']['zshape']
    datashape = record_dict['info']['datashape']
    zsize = reduce(lambda a, b: a * b, zshape, 1)
    datashape_r = (*datashape[:dim], zsize)

    if dim == 1 and zsize < 101:
        x, z = record_dict['data']
        z = z.reshape(datashape_r)
        z = np.abs(z) if np.any(np.iscomplex(z)) else z.real
        for i in range(zsize):
            _ = [a.remove() for a in axes[i].get_lines()] if remove else []
            axes[i].plot(x, z[:, i], 'C0')
            axes[i].plot(x, z[:, i], 'C0.')
    elif dim == 2 and zsize < 101:
        x, y, z = record_dict['data']
        x_step, y_step = x[1] - x[0], y[1] - y[0]
        if transpose:
            x, y = y, x
            x_step, y_step = y_step, x_step
        z = z.reshape(datashape_r)
        z = np.abs(z) if np.any(np.iscomplex(z)) else z.real
        for i in range(zsize):
            _z = z[:, :, i]
            _ = [a.remove() for a in axes[i].get_images()] if remove else []
            if transpose:
                _z = _z.T
            im = axes[i].imshow(_z,
                                extent=(y[0] - y_step / 2, y[-1] + y_step / 2,
                                        x[0] - x_step / 2, x[-1] + x_step / 2),
                                origin='lower',
                                aspect='auto')
            cb_list[i].update_normal(im)  # 更新对应的colorbar
    else:
        pass

    return fig, axes



import copy
import itertools
import logging
import re
import string
from collections import defaultdict
from functools import partial
from typing import Any, Callable, Literal, NamedTuple

log = logging.getLogger(__name__)

Decorator = Callable[[Callable], Callable]

_buildin_set = set


class Source():
    pass


class Sink():
    pass


def action(key: str,
           method: Literal['get', 'set', 'post', 'delete'] = 'get',
           **kwds) -> Decorator:

    if any(c in key for c in ",()[]<>"):
        raise ValueError('Invalid key: ' + key)

    def decorator(func):
        func.__action__ = key, method, kwds
        return func

    return decorator


def get(key: str, **kwds) -> Decorator:
    return action(key, 'get', **kwds)


def set(key: str, **kwds) -> Decorator:
    return action(key, 'set', **kwds)


def post(key: str, **kwds) -> Decorator:
    return action(key, 'post', **kwds)


def delete(key: str, **kwds) -> Decorator:
    return action(key, 'delete', **kwds)


class _Exclusion(NamedTuple):
    keys: list


def exclude(sections: list):
    return _Exclusion(sections)


def _add_action(attrs: dict, key: str, method: str, func: Callable, doc: dict,
                sections: dict) -> None:
    try:
        mapping = attrs[f'__{method}_actions__']
    except KeyError:
        raise ValueError('Invalid method: ' + method)
    arguments = re.findall(r'\{(\w+)\}', key)
    doc[method][key] = func.__doc__
    matrix = {}
    for arg in arguments:
        if (arg not in attrs and arg not in sections or arg in sections
                and arg not in attrs and isinstance(sections[arg], _Exclusion)
                or arg in sections and arg in attrs
                and isinstance(sections[arg], _Exclusion) and len(
                    _buildin_set(attrs[arg]) -
                    _buildin_set(sections[arg].keys)) == 0):
            raise ValueError(
                f'Undefined section: {arg!r} in @action({key!r}, {method!r})')
        if arg in sections and not isinstance(sections[arg], _Exclusion):
            matrix[arg] = sections[arg]
        else:
            if arg in sections:
                matrix[arg] = [
                    k for k in attrs[arg] if k not in sections[arg].keys
                ]
            else:
                matrix[arg] = attrs[arg]
    for values in itertools.product(*[matrix[arg] for arg in arguments]):
        kwds = dict(zip(arguments, values))
        # mapping[key.format(**kwds)] = partial(func, **kwds)
        mapping[string.Template(key).substitute(kwds)] = partial(func, **kwds)


def _build_docs(mapping: dict, attrs: dict) -> str:
    docs = []
    for key, doc in mapping.items():
        if not doc:
            doc = "No documentation."
        docs.append(f"key = \"{key}\"")
        lines = doc.strip().split('\n')
        docs.extend(lines)
        docs.append("")
    return '\n'.join(docs)


class DeviceMeta(type):

    def __new__(cls, name, bases, attrs):
        doc = defaultdict(dict)
        for method in ['get', 'set', 'post', 'delete']:
            attrs.setdefault(f'__{method}_actions__', {})
        for attr in attrs.values():
            if hasattr(attr, '__action__'):
                key, method, kwds = attr.__action__
                _add_action(attrs, key, method, attr, doc, kwds)
        new_class = super().__new__(cls, name, bases, attrs)

        for method in ['get', 'set', 'post', 'delete']:
            getattr(new_class,
                    method).__doc__ = f"{method.upper()}\n\n" + _build_docs(
                        doc[method], attrs)
        return new_class


class BaseDevice(metaclass=DeviceMeta):
    __log__ = None

    def __init__(self, address: str = None, **options):
        self._state = {}
        self._status = "NOT CONNECTED"
        self.address = address
        self.options = options

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    @property
    def log(self):
        if self.__log__ is None:
            self.__log__ = logging.getLogger(
                f"{self.__class__.__module__}.{self.__class__.__name__}")
        return self.__log__

    @property
    def state(self):
        return copy.deepcopy(self._state)
    
    def get_state(self, key: str, default=None) -> Any:
        return self._state.get(key, default)

    @property
    def status(self):
        return self._status

    def open(self) -> None:
        self._status = "CONNECTED"

    def close(self) -> None:
        self._status = "NOT CONNECTED"
        self._state.clear()

    def reset(self) -> None:
        self._state.clear()

    def get(self, key: str, default: Any = None) -> Any:
        self.log.info(f'Get {key!r}')
        if key in self.__get_actions__:
            result = self.__get_actions__[key](self)
            self._state[key] = result
            return result
        else:
            return self._state.get(key, default)

    def set(self, key: str, value: Any = None) -> None:
        self.log.info(f'Set {key!r} = {value!r}')
        self.__set_actions__[key](self, value)
        self._state[key] = value

    def post(self, key: str, value: Any = None) -> Any:
        self.log.info(f'Post {key!r} = {value!r}')
        return self.__post_actions__[key](self, value)

    def delete(self, key: str) -> None:
        self.log.info(f'Delete {key!r}')
        self.__delete_actions__[key](self)
        del self._state[key]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.address!r})'


class VisaDevice(BaseDevice):

    error_query = 'SYST:ERR?'

    def open(self) -> None:
        import pyvisa
        kwds = self.options.copy()
        if 'backend' in kwds:
            rm = pyvisa.ResourceManager(kwds.pop('backend'))
        else:
            rm = pyvisa.ResourceManager()
        self.resource = rm.open_resource(self.address, **kwds)
        self.errors = []
        super().open()

    def close(self) -> None:
        super().close()
        self.resource.close()

    def reset(self) -> None:
        super().reset()
        self.resource.write('*RST')

    def check_error(self):
        if self.error_query:
            while True:
                error = self.resource.query(self.error_query)
                error_code = int(error.split(',')[0])
                if error_code == 0:
                    break
                self.errors.append(error)

    @get('idn')
    def get_idn(self) -> str:
        """Get instrument identification."""
        return self.resource.query('*IDN?')

    @get('opc')
    def get_opc(self) -> bool:
        """Get operation complete."""
        return bool(int(self.resource.query('*OPC?')))

    @get('errors')
    def get_errors(self) -> list[str]:
        """Get error queue."""
        self.check_error()
        errors = self.errors
        self.errors = []
        return errors

    @set('timeout')
    def set_timeout(self, value: float) -> None:
        """Set timeout in seconds."""
        self.resource.timeout = round(value * 1000)

    @get('timeout')
    def get_timeout(self) -> float:
        """Get timeout in seconds."""
        return self.resource.timeout / 1000



import asyncio
import functools
import inspect


def async_command(func=None):
    """
    Decorator to mark a function as an asynchronous command.
    """
    if func is None:
        return async_command

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return run(func, *args, **kwargs)

    return wrapper


def run(main, *args, **kwds):
    if inspect.iscoroutinefunction(main):
        try:
            import uvloop
            uvloop.run(main(*args, **kwds))
        except (ImportError, ModuleNotFoundError):
            asyncio.run(main(*args, **kwds))
    else:
        main(*args, **kwds)



class _Missing():

    def __repr__(self):
        return "Missing"


Missing = _Missing()


def connect_trace(*mappings):
    if not mappings:
        return {}
    result = {}
    first_mapping = mappings[0]
    for key in first_mapping:
        current_value = key
        trajectory = []
        for mapping in mappings:
            if current_value is Missing:
                trajectory.append(Missing)
                continue
            if current_value in mapping:
                next_value = mapping[current_value]
                trajectory.append(next_value)
                current_value = next_value
            else:
                trajectory.append(Missing)
                current_value = Missing
        result[key] = trajectory
    return result


def connect(*mappings):
    if not mappings:
        return {}
    return {
        k: v[-1]
        for k, v in connect_trace(*mappings).items() if v[-1] is not Missing
    }



import atexit
import inspect
import os
import pickle
import shutil
import sys
import tempfile
import time
import warnings
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType
from typing import Any

from loguru import logger

from .registry import Registry
from .storage import Report
from .template import (TemplateKeyError, TemplateTypeError, decode_mapping,
                       inject_mapping)


class SetConfigWorkflow():
    __timeout__ = None
    __mtime__ = 0
    __source__ = ''

    def __init__(self, key):
        self.key = key
        self.__workflow_id__ = f"cfg:{self.key}"

    def depends(self):
        return []

    def check_state(self, history: Report) -> bool:
        reg = Registry()
        try:
            return self._equal(history.parameters[self.key], reg.get(self.key))
        except:
            return False

    def calibrate(self):
        reg = Registry()
        try:
            value = reg.get(self.key)
        except:
            value = eval(input(f'"{self.key}": '))
        return value

    def analyze(self, report: Report, history: Report):
        report.state = 'OK'
        report.parameters = {self.key: report.data}
        return report

    def check(self):
        return self.calibrate()

    def check_analyze(self, report: Report, history: Report | None):
        if self.check_state(history):
            report.state = 'OK'
            report.parameters = {self.key: history.data}
        else:
            report.state = 'Outdated'
        return report

    @staticmethod
    def _equal(a, b):
        import numpy as np

        if a is b:
            return True
        try:
            return a == b
        except:
            pass

        if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
            return a.shape == b.shape and np.all(a == b)

        return False

    def __hash__(self):
        return hash(self.__workflow_id__)


WorkflowType = ModuleType | SetConfigWorkflow


def can_call_without_args(func):
    if not callable(func):
        return False

    # 获取函数签名
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        # 如果有参数没有默认值且不是可变参数，则无法无参调用
        if (param.default is param.empty and param.kind
                not in (param.VAR_POSITIONAL, param.VAR_KEYWORD)):
            return False
    return True


def verify_calibrate_method(module: WorkflowType):
    if not hasattr(module, 'calibrate'):
        raise AttributeError(
            f"Workflow {module.__file__} does not have 'calibrate' function")

    if not can_call_without_args(module.calibrate):
        raise AttributeError(
            f"Workflow {module.__file__} 'calibrate' function should not have any parameters"
        )

    if not hasattr(module, 'analyze'):
        raise AttributeError(
            f"Workflow {module.__file__} does not have 'analyze' function")


def verify_check_method(module: WorkflowType):
    if not hasattr(module, 'check'):
        warnings.warn(
            f"Workflow {module.__file__} does not have 'check' function, it will be set to 'calibrate' function"
        )
    else:
        if not can_call_without_args(module.check):
            raise AttributeError(
                f"Workflow {module.__file__} 'check' function should not have any parameters"
            )

        if not hasattr(module, 'check_analyze'):
            raise AttributeError(
                f"Workflow {module.__file__} has 'check' function but does not have 'check_analyze' function"
            )


def verify_dependence_key(workflow: str | tuple[str, dict[str, Any]]
                          | tuple[str, str, dict[str, Any]], base_path: Path):
    if isinstance(workflow, str):
        return
    if not isinstance(workflow, tuple) or len(workflow) not in [2, 3]:
        raise ValueError(f"Invalid workflow: {workflow}")

    if len(workflow) == 2 and isinstance(workflow[1], str):
        file_name, mapping = workflow
        if not Path(file_name).exists():
            raise FileNotFoundError(f"File not found: {file_name}")
    elif len(workflow) == 2 and isinstance(workflow[1], dict):
        template_path, mapping = workflow
        if not (Path(base_path) / template_path).exists():
            raise FileNotFoundError(f"File not found: {template_path}")
    elif len(workflow) == 3:
        template_path, target_path, mapping = workflow
        if not (Path(base_path) / template_path).exists():
            raise FileNotFoundError(f"File not found: {template_path}")
        if not isinstance(target_path, (Path, str)) or target_path == '':
            raise ValueError(f"Invalid target_path: {target_path}")
        if not isinstance(target_path, (Path, str)):
            raise ValueError(f"Invalid target_path: {target_path}")
        if Path(target_path).suffix != '.py':
            raise ValueError(
                f"Invalid target_path: {target_path}. Only .py file is supported"
            )
    else:
        raise ValueError(f"Invalid workflow: {workflow}")

    if not isinstance(mapping, dict):
        raise ValueError(f"Invalid mapping: {mapping}")

    for key, value in mapping.items():
        if not isinstance(key, str):
            raise ValueError(
                f"Invalid key: {key}, should be str type and valid identifier")
        if not key.isidentifier():
            raise ValueError(f"Invalid key: {key}, should be identifier")
        try:
            pickle.dumps(value)
        except Exception as e:
            raise ValueError(
                f"Invalid value: {key}: {value}, should be pickleable") from e
    return


def verify_depends(module: WorkflowType, base_path):
    if not hasattr(module, 'depends'):
        return

    deps = []

    if callable(module.depends):
        if not can_call_without_args(module.depends):
            raise AttributeError(
                f"Workflow {module.__file__} 'depends' function should not have any parameters"
            )
        deps = list(module.depends())
    elif isinstance(module.depends, (list, tuple)):
        deps = module.depends
    else:
        raise AttributeError(
            f"Workflow {module.__file__} 'depends' should be a callable or a list"
        )
    for workflow in deps:
        verify_dependence_key(workflow, base_path)


def verify_entries(module: WorkflowType, base_path):
    if not hasattr(module, 'entries'):
        return

    deps = []

    if callable(module.entries):
        if not can_call_without_args(module.entries):
            raise AttributeError(
                f"Workflow {module.__file__} 'entries' function should not have any parameters"
            )
        deps = list(module.entries())
    elif isinstance(module.entries, (list, tuple)):
        deps = module.entries
    else:
        raise AttributeError(
            f"Workflow {module.__file__} 'entries' should be a callable or a list"
        )
    for workflow in deps:
        verify_dependence_key(workflow, base_path)


def is_workflow(module: ModuleType) -> bool:
    try:
        verify_calibrate_method(module)
        return True
    except AttributeError:
        return False


def is_template(path: str | Path) -> bool:
    path = Path(path)
    if path.name == 'template.py':
        return True
    if path.name.endswith('_template.py'):
        return True
    if 'templates' in path.parts:
        return True
    return False


def find_unreferenced_workflows(path: str) -> list[str]:
    root = Path(path).resolve()
    workflows = []
    workflow_paths: set[str] = set()

    # Collect all workflow modules
    for file_path in root.rglob("*.py"):
        if file_path.name == "__init__.py":
            continue
        if is_template(file_path):
            continue
        try:
            rel_path = file_path.relative_to(root)
        except ValueError:
            continue

        module = load_workflow_from_file(str(rel_path), root)

        if is_workflow(module):
            rel_str = str(rel_path)
            workflows.append(rel_str)
            workflow_paths.add(rel_str)

    dependencies: set[str] = set()

    # Check dependencies for each workflow module
    for rel_str in workflows:
        module = load_workflow_from_file(rel_str, root)

        depends_func = getattr(module, "depends", None)
        if depends_func and callable(depends_func):
            if not can_call_without_args(depends_func):
                warnings.warn(
                    f"Skipping depends() in {rel_str} as it requires arguments"
                )
                continue
            try:
                depends_list = [
                    n.__workflow_id__ for n in get_dependents(module, root)
                ]
            except Exception as e:
                warnings.warn(f"Error calling depends() in {rel_str}: {e}")
                continue

            if not isinstance(depends_list, list) or not all(
                    isinstance(item, str) for item in depends_list):
                warnings.warn(
                    f"depends() in {rel_str} did not return a list of strings")
                continue

            for dep in depends_list:
                dep_full = (root / dep).resolve()
                try:
                    dep_rel = dep_full.relative_to(root)
                except ValueError:
                    continue
                dep_rel_str = str(dep_rel)
                if dep_rel_str in workflow_paths:
                    dependencies.add(dep_rel_str)

    # Determine unreferenced workflows
    unreferenced = [wp for wp in workflows if wp not in dependencies]
    return unreferenced


def load_workflow_from_file(file_name: str,
                            base_path: str | Path,
                            package='workflows',
                            veryfy_source_code: bool = True) -> WorkflowType:
    base_path = Path(base_path)
    path = Path(file_name)
    if not (base_path / path).exists():
        raise FileNotFoundError(f"File not found: {base_path / path}")
    module_name = f"{package}.{'.'.join([*path.parts[:-1], path.stem])}"
    spec = spec_from_file_location(module_name, base_path / path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    module.__mtime__ = (base_path / path).stat().st_mtime
    source_code = (base_path / path).read_text()
    module.__source__ = source_code

    sys.modules[module_name] = module

    if hasattr(module, 'entries'):
        if veryfy_source_code:
            verify_entries(module, base_path)
        return module

    if not hasattr(module, '__timeout__'):
        module.__timeout__ = None

    if not hasattr(module, 'depends'):
        module.depends = lambda: []

    if veryfy_source_code:
        verify_depends(module, base_path)
        verify_calibrate_method(module)
        verify_check_method(module)

    return module


def load_workflow_from_source_code(workflow_id: str,
                                   source_code: str,
                                   package='workflows') -> WorkflowType:
    workflow_path = Path(workflow_id)
    module_name = f"{package}.{'.'.join([*workflow_path.parts[:-1], workflow_path.stem])}"

    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.py').name
    with open(temp_file, 'w') as f:
        f.write(source_code)

    atexit.register(lambda f=temp_file: Path(f).unlink(missing_ok=True))

    spec = spec_from_file_location(module_name, temp_file)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)

    sys.modules[module_name] = module

    module.__source__ = source_code
    module.__mtime__ = 0
    module.__workflow_id__ = workflow_id

    if not hasattr(module, '__timeout__'):
        module.__timeout__ = None

    if not hasattr(module, 'depends'):
        module.depends = lambda: []

    return module


def _generate_target_file_path(template_path: str | Path, hash_str: str,
                               content: str, base_path: str | Path) -> Path:
    path = Path(template_path)
    if path.stem == 'template':
        path = path.parent / f'tmp{hash_str}.py'
    elif path.stem.endswith('_template'):
        path = path.parent / path.stem.replace('_template',
                                               f'_tmp{hash_str}.py')
    else:
        path = path.parent / f'{path.stem}_tmp{hash_str}.py'

    if 'templates' in path.parts:
        return Path(*['run' if p == 'templates' else p for p in path.parts])
    else:
        return Path('run') / path


def load_workflow_from_template(
        template_path: str,
        mapping: dict[str, str],
        base_path: str | Path,
        target_path: str | None = None,
        package='workflows',
        mtime: float = 0,
        veryfy_source_code: bool = True) -> WorkflowType:
    base_path = Path(base_path)
    path = Path(template_path)

    with open(base_path / path) as f:
        template = f.read()

    mtime = max((base_path / template_path).stat().st_mtime, mtime)

    content, hash_str = inject_mapping(template, mapping, str(path))

    if target_path is None:
        path = _generate_target_file_path(template_path, hash_str, content,
                                          base_path)
    else:
        path = target_path

    file = base_path / path
    if not file.exists() or (file.stat().st_mtime < mtime
                             and file.read_text() != content):
        file.parent.mkdir(parents=True, exist_ok=True)
        with open(file, 'w') as f:
            f.write(content)
    elif file.read_text() != content:
        logger.warning(
            f"`{file}` already exists and is different from the new one generated from template `{template_path}`"
        )

    module = load_workflow_from_file(str(path), base_path, package,
                                     veryfy_source_code)
    module.__mtime__ = max(mtime, module.__mtime__)
    if module.__source__ == content:
        module.__source__ = template, mapping, str(template_path)

    return module


def load_workflow(workflow: str | tuple[str, dict],
                  base_path: str | Path,
                  package='workflows',
                  mtime: float = 0,
                  inject: dict | None = None,
                  veryfy_source_code: bool = True) -> WorkflowType:
    if isinstance(workflow, tuple):
        if len(workflow) == 2:
            file_name, mapping = workflow
            if inject is None:
                w = load_workflow_from_template(file_name, mapping, base_path,
                                                None, package, mtime,
                                                veryfy_source_code)
            else:
                w = load_workflow_from_template(file_name, inject, base_path,
                                                None, package, mtime,
                                                veryfy_source_code)
        elif len(workflow) == 3:
            template_path, target_path, mapping = workflow
            if inject is None:
                w = load_workflow_from_template(template_path, mapping,
                                                base_path, target_path,
                                                package, mtime,
                                                veryfy_source_code)
            else:
                w = load_workflow_from_template(template_path, inject,
                                                base_path, target_path,
                                                package, mtime,
                                                veryfy_source_code)
        else:
            raise ValueError(f"Invalid workflow: {workflow}")
        w.__workflow_id__ = str(Path(w.__file__).relative_to(base_path))
    elif isinstance(workflow, str):
        if workflow.startswith('cfg:'):
            key = workflow[4:]
            w = SetConfigWorkflow(key)
            w.__workflow_id__ = workflow
        else:
            w = load_workflow_from_file(workflow, base_path, package,
                                        veryfy_source_code)
            w.__workflow_id__ = str(Path(w.__file__).relative_to(base_path))
    else:
        raise TypeError(f"Invalid workflow: {workflow}")

    return w


def _load_workflow_list(workflow, lst, code_path, veryfy_source_code):
    ret = []
    for i, n in enumerate(lst):
        try:
            ret.append(
                load_workflow(n,
                              code_path,
                              mtime=workflow.__mtime__,
                              veryfy_source_code=veryfy_source_code))
        except TemplateKeyError as e:
            msg, *_ = e.args
            e.args = (
                f"Workflow {workflow.__workflow_id__} entry {i}: {msg}", )
            raise e
    return ret


def get_dependents(workflow: WorkflowType, code_path: str | Path,
                   veryfy_source_code: bool) -> list[WorkflowType]:
    if callable(getattr(workflow, 'depends', None)):
        if not can_call_without_args(workflow.depends):
            raise AttributeError(
                f'Workflow {workflow.__workflow_id__} "depends" function should not have any parameters'
            )
        return _load_workflow_list(workflow, workflow.depends(), code_path,
                                   veryfy_source_code)
    elif isinstance(getattr(workflow, 'depends', None), (list, tuple)):
        return _load_workflow_list(workflow, workflow.depends, code_path,
                                   veryfy_source_code)
    elif getattr(workflow, 'depends', None) is None:
        return []
    else:
        raise AttributeError(
            f'Workflow {workflow.__workflow_id__} "depends" should be a callable or a list'
        )


def get_entries(workflow: WorkflowType, code_path: str | Path,
                veryfy_source_code: bool) -> list[WorkflowType]:
    if callable(getattr(workflow, 'entries', None)):
        if not can_call_without_args(workflow.entries):
            raise AttributeError(
                f'Workflow {workflow.__workflow_id__} "entries" function should not have any parameters'
            )
        return _load_workflow_list(workflow, workflow.entries(), code_path,
                                   veryfy_source_code)
    elif isinstance(getattr(workflow, 'entries', None), (list, tuple)):
        return _load_workflow_list(workflow, workflow.entries, code_path,
                                   veryfy_source_code)
    elif getattr(workflow, 'entries', None) is None:
        return []
    else:
        raise AttributeError(
            f'Workflow {workflow.__workflow_id__} "entries" should be a callable or a list'
        )


def make_graph(workflow: WorkflowType,
               graph: dict,
               code_path: str | Path,
               veryfy_source_code: bool = True):
    if workflow.__workflow_id__ in graph:
        return graph
    graph[workflow.__workflow_id__] = []

    if hasattr(workflow, 'entries'):
        for w in get_entries(workflow, code_path, veryfy_source_code):
            graph[workflow.__workflow_id__].append(w.__workflow_id__)
            make_graph(w,
                       graph=graph,
                       code_path=code_path,
                       veryfy_source_code=veryfy_source_code)
    elif hasattr(workflow, 'depends'):
        for w in get_dependents(workflow, code_path, veryfy_source_code):
            graph[workflow.__workflow_id__].append(w.__workflow_id__)
            make_graph(w,
                       graph=graph,
                       code_path=code_path,
                       veryfy_source_code=veryfy_source_code)
    if graph[workflow.__workflow_id__] == []:
        del graph[workflow.__workflow_id__]

    return graph



import asyncio
import os
from urllib.parse import quote_plus

from fastapi import FastAPI, Request
from fastapi.responses import Response
from pydantic import BaseModel

NGINX_DIR = 'C:/Users/Administrator/Desktop/nginx/nginx-1.20.2'
HOST_URL = 'https://systemq.baqis.ac.cn'

db = {
    "n01": {
        "notebook_port": 1000,
        "codeserver_port": 1001,
        "ip": "127.0.0.1",
        "token": "ed747047-1d9d-47b0-8fc2-b4914f8a9bd9",
    },
    "n02": {
        "notebook_port": 1002,
        "codeserver_port": 1003,
        "ip": "127.0.0.1",
        "token": "ed423dd1-05fb-48d2-a720-dbebfb2ea3e1",
    },
    "n03": {
        "notebook_port": 1004,
        "codeserver_port": 1005,
        "ip": "127.0.0.1",
        "token": "e3b0c442-98fc-11e4-8dfc-aa07a5b093d3",
    },
}


def fmt_config(ip, notebook_port, codeserver_port, local_notebook_port,
               local_codeserver_port, page):
    return """
server {""" + f"""
    listen {notebook_port} ssl;
    listen [::]:{notebook_port} ssl;
""" + """
    include ssl/ssl.conf;

    server_name  systemq.baqis.ac.cn;

    client_max_body_size 100M;

    # notebook
    location / {
        include auth/auth.conf;
""" + f"""
        proxy_pass http://{ip}:{local_notebook_port}/;""" + """
        proxy_set_header X-Real_IP $remote_addr;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-NginX-Proxy true;
        proxy_ssl_session_reuse off;
        proxy_set_header Host $http_host;

        proxy_redirect off;
        proxy_http_version 1.1;
        proxy_set_header Upgrade $http_upgrade;
        proxy_set_header Connection "upgrade";
    }

    location /static/ {
        alias  html/_jupyter_notebook/static/;
    }

    location /login {
        include auth/auth.conf;
        default_type text/html;
        add_header Content-Type "text/html; charset=UTF-8";
""" + f"""
        return 200 "{page}";
""" + """
    }

    include auth/location.conf;
}

server {""" + f"""
    listen  {codeserver_port} ssl;
    listen [::]:{codeserver_port} ssl;
""" + """
    include ssl/ssl.conf;

    server_name  systemq.baqis.ac.cn;

    client_max_body_size 100M;

    # vscode
    location / {
        include auth/auth.conf;
""" + f"""
        proxy_pass http://{ip}:{local_codeserver_port}/;""" + """
        proxy_set_header X-Real_IP $remote_addr;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-NginX-Proxy true;
        proxy_ssl_session_reuse off;
        proxy_set_header Host $http_host;

        proxy_redirect off;
        proxy_http_version 1.1;
        proxy_set_header Upgrade $http_upgrade;
        proxy_set_header Connection "upgrade";
    }

    location /oss-dev/static/out/ {
        root   html;
    }

    include auth/location.conf;
}
"""


def fmt_page(notebook_url, codeserver_url):
    return """
<!DOCTYPE html>
<html>
<head>
<title>Welcome to systemq!</title>
<style>
body {
    width: 35em;
    margin: 0 auto;
    font-family: Tahoma, Verdana, Arial, sans-serif;
}
</style>
</head>
<body>
<h1>Welcome to systemq!</h1>""" + f"""
<a href=\\"{notebook_url}\\">notebook</a>.<br/>
<a href=\\"{codeserver_url}\\">codeserver</a>.<br/>
</body>
</html>
"""


class Node(BaseModel):
    token: str
    name: str
    notebook_port: int = 8888
    codeserver_port: int = 14263
    codeserver_path: str = ""


app = FastAPI()


@app.post("/auth")
async def auth(request: Request, node: Node):
    try:
        assert node.name in db
        assert node.token == db[node.name]['token']
        notebook_port = db[node.name]['notebook_port']
        codeserver_port = db[node.name]['codeserver_port']

        ip = request.client.host
        if ip == db[node.name]['ip']:
            return Response(status_code=200)
        else:
            print("    CHANGE IP", db[node.name]['ip'], "===>", ip)
            db[node.name]['ip'] = ip

        #ip = request.headers.get('x-real-ip')
        local_notebook_port = node.notebook_port
        local_codeserver_port = node.codeserver_port
        codeserver_path = quote_plus(node.codeserver_path)

        codeserver_url = f"{HOST_URL}:{codeserver_port}?tkn=lZBAQISFF532&folder={codeserver_path}"
        notebook_url = f"{HOST_URL}:{notebook_port}/tree"

        page = fmt_page(notebook_url, codeserver_url)
        page = ''.join(page.splitlines())

        config = fmt_config(ip, notebook_port, codeserver_port,
                            local_notebook_port, local_codeserver_port, page)

        with open(f"{NGINX_DIR}/conf/servers/{node.name}.conf", "w") as f:
            f.write(config)

        await asyncio.sleep(0.1)

        cwd = os.getcwd()

        os.chdir(NGINX_DIR)

        os.system(f'{NGINX_DIR}/nginx.exe -s reload')

        os.chdir(cwd)

        return Response(status_code=200)
    except:
        return Response(status_code=400)



import functools
import graphlib
import importlib
import os
import sys
from pathlib import Path

import click
import rich
from loguru import logger

from ..cli.config import get_config_value, log_options
from ..cli.decorators import async_command
from .load import (WorkflowType, find_unreferenced_workflows, get_entries,
                   load_workflow, make_graph)
from .registry import Registry, set_config_api
from .schedule import CalibrationFailedError
from .schedule import maintain as maintain_workflow
from .schedule import run as run_workflow
from .utils import workflow_template


@logger.catch(reraise=True)
def run_script(script_path):
    """Run a script in a new terminal."""
    import subprocess
    import sys

    proc = subprocess.Popen([sys.executable, script_path])
    proc.communicate()


@logger.catch(reraise=True)
def check_toplogy(workflow: WorkflowType,
                  code_path: str | Path,
                  veryfy_source_code: bool = True) -> dict:
    graph = {}
    try:
        graphlib.TopologicalSorter(
            make_graph(workflow,
                       graph,
                       code_path,
                       veryfy_source_code=veryfy_source_code)).static_order()
    except graphlib.CycleError as e:
        logger.error(
            f"Workflow {workflow.__workflow_id__} has a circular dependency: {e}"
        )
        raise e
    return graph


def command_option(command_name):
    """命令专属配置装饰器工厂"""

    def decorator(func):

        @click.option(
            '--code',
            '-c',
            default=lambda: get_config_value("code", str, command_name),
            help='The path of the code.')
        @click.option(
            '--data',
            '-d',
            default=lambda: get_config_value("data", str, command_name),
            help='The path of the data.')
        @click.option(
            '--api',
            '-a',
            default=lambda: get_config_value("api", str, command_name),
            help='The modlule name of the api.')
        @click.option(
            '--bootstrap',
            '-b',
            default=lambda: get_config_value("bootstrap", Path, command_name),
            help='The path of the bootstrap.')
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if 'code' in kwargs and kwargs['code'] is not None:
                code = os.path.expanduser(kwargs['code'])
                if code not in sys.path:
                    sys.path.insert(0, code)
            bootstrap = kwargs.pop('bootstrap')
            if bootstrap is not None:
                run_script(bootstrap)
            return func(*args, **kwargs)

        return wrapper

    return decorator


@click.command()
@click.argument('workflow')
@click.option('--code',
              '-c',
              default=lambda: get_config_value("code", str, 'create'),
              help='The path of the code.')
@log_options('create')
def create(workflow, code):
    """
    Create a new workflow file.
    """
    logger.info(f'[CMD]: create {workflow} --code {code}')
    if code is None:
        code = Path.cwd()

    fname = Path(code) / f'{workflow}'
    fname = Path(os.path.expanduser(fname))
    if fname.exists():
        click.echo(f'{workflow} already exists')
        return

    fname.parent.mkdir(parents=True, exist_ok=True)
    deps = find_unreferenced_workflows(code)

    with open(fname, 'w') as f:
        f.write(workflow_template(workflow, list(deps)))
    click.echo(f'{workflow} created')


@click.command()
@click.argument('key')
@click.argument('value', type=str)
@click.option('--api',
              '-a',
              default=lambda: get_config_value("api", str, 'set'),
              help='The modlule name of the api.')
@log_options('set')
def set(key, value, api):
    """
    Set a config.
    """
    logger.info(f'[CMD]: set {key} {value} --api {api}')
    reg = Registry()
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)
    try:
        value = eval(value)
    except:
        pass
    reg.set(key, value)


@click.command()
@click.argument('key')
@click.option('--api',
              '-a',
              default=lambda: get_config_value("api", str, 'get'),
              help='The modlule name of the api.')
@log_options('get')
def get(key, api):
    """
    Get a config.
    """
    logger.info(f'[CMD]: get {key} --api {api}')
    reg = Registry()
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)
    rich.print(reg.get(key))


@click.command()
@click.argument('file')
@click.option('--api',
              '-a',
              default=lambda: get_config_value("api", str, 'get'),
              help='The modlule name of the api.')
@click.option('--format',
              '-f',
              default='json',
              help='The format of the config.')
@log_options('export')
def export(file, api, format):
    """
    Export a config.
    """
    logger.info(f'[CMD]: export {file} --api {api}')
    reg = Registry()
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)
    cfg = reg.export()
    if format == 'json':
        import json
        with open(file, 'w') as f:
            json.dump(cfg, f, indent=4)
    elif format == 'yaml':
        import yaml
        with open(file, 'w') as f:
            yaml.dump(cfg, f)
    elif format == 'pickle':
        import pickle
        with open(file, 'wb') as f:
            pickle.dump(cfg, f)
    else:
        raise ValueError(f'Unknown format: {format}')


@click.command()
@click.argument('file')
@click.option('--api',
              '-a',
              default=lambda: get_config_value("api", str, 'get'),
              help='The modlule name of the api.')
@click.option('--format',
              '-f',
              default='json',
              help='The format of the config.')
@log_options('load')
def load(file, api, format):
    """
    Load a config.
    """
    logger.info(f'[CMD]: load {file} --api {api}')
    reg = Registry()
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)
    if format == 'json':
        import json
        with open(file, 'r') as f:
            cfg = json.load(f)
    elif format == 'yaml':
        import yaml
        with open(file, 'r') as f:
            cfg = yaml.load(f, Loader=yaml.FullLoader)
    elif format == 'pickle':
        import pickle
        with open(file, 'rb') as f:
            cfg = pickle.load(f)
    else:
        raise ValueError(f'Unknown format: {format}')
    reg.clear()
    reg.update(cfg)


@click.command()
@click.option('--bootstrap',
              '-b',
              default=lambda: get_config_value("bootstrap", Path),
              help='The path of the bootstrap.')
def boot(bootstrap):
    """
    Run a bootstrap script.
    """
    if bootstrap is not None:
        run_script(bootstrap)


@click.command()
@click.argument('workflow')
@click.option('--plot', '-p', is_flag=True, help='Plot the report.')
@click.option('--no-dependents',
              '-n',
              is_flag=True,
              help='Do not run dependents.')
@click.option('--retry', '-r', default=1, type=int, help='Retry times.')
@click.option('--freeze', is_flag=True, help='Freeze the config table.')
@click.option('--fail-fast',
              '-f',
              is_flag=True,
              help='Fail immediately on first error.')
@click.option('--veryfy-source-code',
              is_flag=True,
              help='Veryfy the source code.')
@log_options('run')
@command_option('run')
@async_command
async def run(workflow,
              code,
              data,
              api,
              plot,
              no_dependents,
              retry,
              freeze,
              fail_fast,
              veryfy_source_code=True):
    """
    Run a workflow.

    If the workflow has entries, run all entries.
    If `--no-dependents` is set, only run the workflow itself.
    If `--retry` is set, retry the workflow when calibration failed.
    If `--freeze` is set, freeze the config table.
    If `--plot` is set, plot the report.
    If `--api` is set, use the api to get and update the config table.
    If `--code` is not set, use the current working directory.
    If `--data` is not set, use the `logs` directory in the code path.
    """
    logger.info(
        f'[CMD]: run {workflow} --code {code} --data {data} --api {api}'
        f'{" --plot" if plot else ""}'
        f'{" --no-dependents" if no_dependents else ""}'
        f' --retry {retry}'
        f'{" --freeze " if freeze else ""}')
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)
    if code is None:
        code = Path.cwd()
    if data is None:
        data = Path(code) / 'logs'

    code = Path(os.path.expanduser(code))
    data = Path(os.path.expanduser(data))

    wf = load_workflow(workflow, code, veryfy_source_code=veryfy_source_code)
    check_toplogy(wf, code, veryfy_source_code=veryfy_source_code)

    for i in range(retry):
        try:
            if no_dependents:
                if hasattr(wf, 'entries'):
                    exceptions = []
                    for entry in get_entries(
                            wf, code, veryfy_source_code=veryfy_source_code):
                        try:
                            await run_workflow(
                                entry,
                                code,
                                data,
                                plot=plot,
                                freeze=freeze,
                            )
                        except Exception as e:
                            if fail_fast:
                                raise e
                            exceptions.append(e)
                    if any(exceptions):
                        raise exceptions[0]
                else:
                    await run_workflow(
                        wf,
                        code,
                        data,
                        plot=plot,
                        freeze=freeze,
                    )
            else:
                if hasattr(wf, 'entries'):
                    exceptions = []
                    for entry in get_entries(
                            wf, code, veryfy_source_code=veryfy_source_code):
                        try:
                            await maintain_workflow(
                                entry,
                                code,
                                data,
                                run=True,
                                plot=plot,
                                freeze=freeze,
                                fail_fast=fail_fast,
                                veryfy_source_code=veryfy_source_code,
                            )
                        except Exception as e:
                            if fail_fast:
                                raise e
                            exceptions.append(e)
                    if any(exceptions):
                        raise exceptions[0]
                else:
                    await maintain_workflow(
                        wf,
                        code,
                        data,
                        run=True,
                        plot=plot,
                        freeze=freeze,
                        fail_fast=fail_fast,
                        veryfy_source_code=veryfy_source_code,
                    )
            break
        except CalibrationFailedError as e:
            if i == retry - 1:
                raise e
            logger.warning(f'Calibration failed, retrying ({i + 1}/{retry})')
            continue


@click.command()
@click.argument('workflow')
@click.option('--retry', '-r', default=1, type=int, help='Retry times.')
@click.option('--plot', '-p', is_flag=True, help='Plot the report.')
@click.option('--fail-fast',
              '-f',
              is_flag=True,
              help='Fail immediately on first error.')
@click.option('--veryfy-source-code',
              is_flag=True,
              help='Veryfy the source code.')
@log_options('maintain')
@command_option('maintain')
@async_command
async def maintain(workflow,
                   code,
                   data,
                   api,
                   retry,
                   plot,
                   fail_fast,
                   veryfy_source_code=True):
    """
    Maintain a workflow.

    If the workflow has entries, run all entries.
    If `--retry` is set, retry the workflow when calibration failed.
    If `--plot` is set, plot the report.
    If `--api` is set, use the api to get and update the config table.
    If `--code` is not set, use the current working directory.
    If `--data` is not set, use the `logs` directory in the code path.
    """
    logger.info(
        f'[CMD]: maintain {workflow} --code {code} --data {data} --api {api}'
        f' --retry {retry}'
        f'{" --plot" if plot else ""}')
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)
    if code is None:
        code = Path.cwd()
    if data is None:
        data = Path(code) / 'logs'

    code = Path(os.path.expanduser(code))
    data = Path(os.path.expanduser(data))

    wf = load_workflow(workflow, code, veryfy_source_code=veryfy_source_code)
    check_toplogy(wf, code, veryfy_source_code=veryfy_source_code)

    for i in range(retry):
        try:
            if hasattr(wf, 'entries'):
                exceptions = []
                for entry in get_entries(
                        wf, code, veryfy_source_code=veryfy_source_code):
                    try:
                        await maintain_workflow(
                            entry,
                            code,
                            data,
                            run=False,
                            plot=plot,
                            freeze=False,
                            fail_fast=fail_fast,
                            veryfy_source_code=veryfy_source_code,
                        )
                    except Exception as e:
                        if fail_fast:
                            raise e
                        exceptions.append(e)
                if any(exceptions):
                    raise exceptions[0]
            else:
                await maintain_workflow(
                    wf,
                    code,
                    data,
                    run=False,
                    plot=plot,
                    freeze=False,
                    fail_fast=fail_fast,
                    veryfy_source_code=veryfy_source_code,
                )
            break
        except CalibrationFailedError as e:
            if i == retry - 1:
                raise e
            logger.warning(f'Calibration failed, retrying ({i + 1}/{retry})')
            continue


@click.command()
@click.argument('report_id')
@click.option('--plot', '-p', is_flag=True, help='Plot the report.')
@log_options('reproduce')
@command_option('reproduce')
@async_command
async def reproduce(report_id, code, data, api, plot):
    """
    Reproduce a report.

    If `--plot` is set, plot the report.
    If `--api` is set, use the api to get and update the config table.
    If `--code` is not set, use the current working directory.
    If `--data` is not set, use the `logs` directory in the code path.
    """
    logger.info(
        f'[CMD]: reproduce {report_id} --code {code} --data {data} --api {api}'
        f'{" --plot" if plot else ""}')
    if api is not None:
        api = importlib.import_module(api)
        set_config_api(api.query_config, api.update_config, api.delete_config,
                       api.export_config, api.clear_config)
    if code is None:
        code = Path.cwd()
    if data is None:
        data = Path(code) / 'logs'

    code = Path(os.path.expanduser(code))
    data = Path(os.path.expanduser(data))

    from .load import load_workflow_from_source_code
    from .storage import get_report_by_index

    reg = Registry()

    r = get_report_by_index(int(report_id), data)

    wf = load_workflow_from_source_code(r.workflow, r.script)
    cfg = reg.export()
    reg.clear()
    reg.update(r.config)
    await run_workflow(wf, code, data, plot=plot, freeze=True)
    reg.clear()
    reg.update(cfg)
